'use strict'

const request = require('client-request')
const lodash = require('lodash')
const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

module.exports = init

// This should be more than enough time
const REQUEST_TIMEOUT = 30000

function init (context) {
  return new EmailAction(context)
}

class EmailAction {
  constructor (context) {
    this._context = context
    this.type = 'emails'
    this._emailEndpoint = this._context.isStaging
      ? 'https://staging-services.nodesource.io/notification'
      : 'https://services.nodesource.io/notification'
  }

  run (params, cb) {
    // find action configuration
    params.actionConfig = this.getActionConfig(params)
    if (!params.actionConfig) return setImmediate(cb)

    const emails = this.filterEmails(params)
    if (!emails || !emails.length) return setImmediate(cb)

    const config = {
      notificationEndpoint: {
        url: this._emailEndpoint
      },
      license: {
        key: this._context.provider.license.key
      },
      httpsAgent: this._context.httpsAgent
    }

    const payload = {
      type: params.event,
      method: 'email',
      data: lodash.defaults({
        recipients: emails,
        savedView: params.view,
        agents: params.agents,
        artifacts: params.artifacts
      }, params.data)
    }

    sendEmails(config, payload, (err, res) => {
      Logger.debug('email results', err, res)
      cb(null, res)
    })
  }

  getActionConfig (params) {
    return params.view.actions.find(action => {
      return action && action.type === 'email' && !!action.emails
    })
  }

  filterEmails (params) {
    let emails = params.actionConfig.emails.split(/, */)

    switch (params.event) {
      case 'nsolid-agent-exit':
        const exitType = params.view.exitType || 'all'
        const hasStack = !!(params.data && params.data.error && params.data.error.stack)

        if (params.data.exitCode === 0) {
          if (exitType === 'abnormal') return []
        } else {
          if (exitType === 'normal') return []
        }

        if (params.view.onlyExceptions && !hasStack) {
          return []
        }

        return emails
      default:
        return emails
    }
  }
}

function sendEmails (config, payload, cb) {
  const url = config.notificationEndpoint.url

  Logger.debug(`emailing via ${url} using license ${config.license.key}`, payload)

  const reqParams = {
    uri: url,
    method: 'POST',
    body: JSON.stringify(payload),
    timeout: REQUEST_TIMEOUT,
    agent: config.httpsAgent,
    headers: {
      'x-nsolid-license-key': config.license.key
    }
  }

  request(reqParams, (err, res) => {
    if (err) {
      Logger.warn(`Error sending email notification: ${err.message}`)
    }
    Logger.debug('email results', err, res.statusCode)
    cb(null, res.statusCode)
  })
}
