'use strict'

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

const messages = {
  'nsolid-saved-view-match': savedViewMatch,
  'nsolid-agent-found': agentFound,
  'nsolid-agent-exit': agentExit,
  'nsolid-snapshot-generated': snapshotGenerated,
  'nsolid-profile-generated': profileGenerated,
  'nsolid-process-blocked': processBlocked,
  'nsolid-vulnerability-detected': vulnerabilityDetected
}

module.exports = function buildSlackMessage (params) {
  if (!messages[params.type]) {
    Logger.error(`No such slack message ${params.type}.`)
    return
  }

  return messages[params.type](params)
}

const entities = {
  '<': '&lt;',
  '>': '&gt;'
}
function escapeChevrons (text) {
  return text.replace(/([<>])/g, c => entities[c])
}

function vulnerabilityDetected (params) {
  const title = `${params.data.vulnerabilities.length > 1 ? 'Vulnerabilities' : 'Vulnerability'} found in process ${describeAgent(params.consoleURL, params.agents[0])}`
  const body = params.data.vulnerabilities.map(vuln => `* ${vuln.package}@${vuln.vulnerable} - ${vuln.title}. More information: <${vuln.url}>`).join('\n')
  return formatRichMessage(title, body, '#ff6040')
}

function processBlocked (params) {
  const time = params.data.blockedFor ? params.data.blockedFor : `at least ${params.data.threshold}`
  const title = `Process ${describeAgent(params.consoleURL, params.agents[0])} has had its event loop blocked for ${time} ms.`
  const body = (params.data && params.data && params.data.stack) ? title + '\n```\n' + params.data.stack + '\n```' : ''
  return formatRichMessage(null, '', '#2d81bd', body, title)
}

function agentExit (params) {
  let title = `Process ${describeAgent(params.consoleURL, params.agents[0])} has exited`
  let body = ''
  if (params.data && params.data.hasOwnProperty('exitCode')) {
    body = `Exit code: ${params.data.exitCode}`
  }

  let content
  if (params.data && params.data.error && params.data.error.stack) {
    content = title + '\n```\n' + params.data.error.stack + '\n```'
    title = null
  }

  if (params.artifacts && params.artifacts.length && params.artifacts[0].url) {
    body += `\nProfile saved at time of exit can be viewed <${params.artifacts[0].url}|here>`
  }

  return formatRichMessage(title, body, '#ffb726', content)
}

function agentFound (params) {
  const title = `Process ${describeAgent(params.consoleURL, params.agents[0])} has been found.`
  return formatRichMessage(title, '', '#ff6040')
}

function snapshotGenerated (params) {
  const title = `New heap snapshot taken of process ${describeAgent(params.consoleURL, params.agents[0])}.`
  const body = `<${params.artifacts[0].url}|Open In Console>`
  return formatRichMessage(title, body, '#4cb5ff')
}

function profileGenerated (params) {
  const title = `New CPU profile taken of process ${describeAgent(params.consoleURL, params.agents[0])}.`
  const body = `<${params.artifacts[0].url}|Open In Console>`
  return formatRichMessage(title, body, '#baff94')
}

function savedViewMatch (params) {
  const title = `${formatProcesses(params.view.consoleURL, params.agents)} entered the "${params.view.name}" saved view by meeting the following criteria:`

  let body = `${formatTerms(params.view.expression, params.agents[0])}\n`
  body += `<${params.view.url}|Open '${escapeChevrons(params.view.name)}' View>\n\n`
  body += `${formatProcessesList(params.view.consoleURL, params.agents)}\n`

  const artifactsLength = params.artifacts.length
  if (artifactsLength > 0) {
    body +=
      'We\'ve created ' +
      (artifactsLength === 1 ? 'an artifact' : 'some artifacts') +
      ' that may help you identify the problem.\n\n'
    for (var i = 0; i < artifactsLength; i++) {
      var artifact = params.artifacts[i]
      switch (artifact.type) {
        case 'snapshot':
          body += '* Heap Snapshot: ' + artifact.url + '\n'
          break
        case 'profile':
          body += '* Profile: ' + artifact.url + '\n'
          break
        default:
          break
      }
    }
  }

  return formatRichMessage(title, body, '#3a824e')
}

function formatTerms (terms, agent) {
  terms = terms.toJSON()
  if (!terms || typeof terms[Symbol.iterator] !== 'function') {
    throw new Error('terms is bad ' + JSON.stringify(terms))
  }
  const lines = []
  for (let term of terms) {
    // > CPU: 125% (> 50%)
    lines.push(`> ${term.field}: ${formatValue(agent, term.field)}${formatOperator(term.operator, term.value || term.values)}`)
  }

  return lines.join('\n')
}

function formatValue (agent, field) {
  if (agent.metrics && agent.metrics[field]) return agent.metrics[field]
  if (agent.info && agent.info[field]) return agent.info[field]
  if (agent[field]) return agent[field]
  return `unknown ${field}`
}

function formatProcesses (url, agents) {
  if (agents.length === 1) {
    return `Process ${describeAgent(url, agents[0])} has`
  } else {
    return `${agents.length} processes have`
  }
}

function formatProcessesList (url, agents) {
  if (agents.length === 1) return ''
  const idList = agents.map(agent => `<${url}/processes/${formatValue(agent, 'id').substr(0, 7)}/|${formatValue(agent, 'id').substr(0, 7)}>`).join(', ')
  return `\nProcesses:\n${idList}\n`
}

function formatOperator (operator, val) {
  if (Array.isArray(val) && val.length === 1) {
    val = val[0]
  }

  const valIsArray = Array.isArray(val)
  switch (operator) {
    case '!~':
      return valIsArray ? ` (does not contain any of "${val.join('", "')}")` : ` (does not contain "${val}")`
    case '~':
      return valIsArray ? ` (contains one of "${val.join('", "')}")` : ` (contains "${val}")`
    case '!=':
      return valIsArray ? ` (not one of "${val.join('", "')}")` : ` (not ${val})`
    case '<':
      return ` (less than ${val})`
    case '<=':
      return ` (less than or equal to ${val})`
    case '>':
      return ` (greater than ${val})`
    case '>=':
      return ` (greater than than or equal to ${val})`
    case '=':
      // If singular equality, don't repeat ourselves by writing "app" is "app"
      return valIsArray ? ` (is one of "${val.join('", "')}")` : ''
  }
}

function describeAgent (url, agent) {
  // Deal with undefined agent, which should not happen, but in case a late
  // (post agent exit) lookup results in an undefined agent, we need to do
  // something here
  return `<${url}/processes/${formatValue(agent, 'id').substr(0, 7)}/|${formatValue(agent, 'id').substr(0, 7)}> (${formatValue(agent, 'app')} on ${formatValue(agent, 'hostname')})`
}

function formatRichMessage (title, body, color, payload, fallback) {
  const message = {
    attachments: [
      {
        fallback: fallback || title,
        color,
        pretext: title,
        text: body,
        footer: 'N|Solid',
        ts: Math.floor(Date.now() / 1000)
      }
    ]
  }
  if (payload != null) {
    message.text = payload
  }
  return message
}
