'use strict'

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)
const request = require('client-request')
const lodash = require('lodash')
const asynclib = require('async')

module.exports = init

// This should be more than enough time
const REQUEST_TIMEOUT = 30000

function init (context) {
  return new WebhookAction(context)
}

class WebhookAction {
  constructor (context) {
    this._context = context
    this.type = 'webhook'
  }

  run (params, cb) {
    // find action configurations
    const actionConfigs = this.getActionConfigs(params)
    if (!actionConfigs || actionConfigs.length === 0) return setImmediate(cb)

    // Don't force these to run in series because one might be slow, but don't
    // allow unlimited parallelism
    asynclib.mapLimit(actionConfigs, 3, (actionConfig, done) => {
      if (!this.shouldRun(params, actionConfig)) return done()

      Logger.debug(`Sending webhook for ${params.view.name || params.event} to ${actionConfig.url}`)
      const webhookParams = lodash.defaults({
        time: new Date(params.now),
        event: params.event,
        agents: params.agents,
        config: params.view,
        assets: params.artifacts
      }, params.data)

      const reqParams = {
        uri: actionConfig.url,
        headers: {
          'Content-Type': 'application/json'
        },
        method: 'POST',
        body: webhookParams,
        timeout: REQUEST_TIMEOUT
      }

      request(reqParams, (err, res) => {
        if (err) {
          Logger.warn(`Error calling webhook at ${actionConfig.url}: ${err.message}`)
        }
        Logger.debug('webhook results', err, res.statusCode)
        done(null, res.statusCode)
      })
    }, (err, codes) => {
      cb(err, codes)
    })
  }

  shouldRun (params, actionConfig) {
    switch (params.event) {
      case 'nsolid-agent-exit':
        const exitType = params.view.exitType || 'all'
        const hasStack = !!(params.data && params.data.error && params.data.error.stack)

        if (params.data.exitCode === 0) {
          if (exitType === 'abnormal') return false
        } else {
          if (exitType === 'normal') return false
        }

        if (params.view.onlyExceptions && !hasStack) {
          return false
        }

        return true
      default:
        return true
    }
  }

  getActionConfigs (params) {
    return params.view.actions.filter(action => {
      return action && action.type === 'webhook' && !!action.url
    })
  }
}
