'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const TimedCache = require('./timed-cache')
const Report = require('./agent-manager/report').Report
const RevocableEventEmitter = require('./RevocableEventEmitter')
const Agent = require('./agent')

const LIVE_AGENT_TTL = 20 * 1000 // 20 seconds to let an agent not respond
const DEAD_AGENT_TTL = 60 * 1000 // 1 minute in ms
const CHECK_INTERVAL = 10 * 1000 // 10 second interval to garbage collect
const LOG_INTERVAL = 30 * 1000   // 30 seconds

exports.create = create

function create (context) {
  return new AgentManager(context)
}

// Models a collection of N|Solid processes.
class AgentManager extends RevocableEventEmitter {
  constructor (context) {
    super()

    this._context = context
    // Note: TimedCache constructor can throw if the ttl argument is invalid
    this._agents = new TimedCache(LIVE_AGENT_TTL, CHECK_INTERVAL)
    this._deadAgents = new TimedCache(DEAD_AGENT_TTL, CHECK_INTERVAL)
    this._commandCBs = new Map()
    this._nextCommandID = util.sequencer()
    this._totalAgentLifetime = 0
    this._totalAgentActiveMs = 0
    this.report = new Report(this)

    // set a listener to update agent vulns when package vulns have been updated
    setImmediate(() => {
      context.eventManager.on('package-vulnerabilities-updated', () => this.updateVulns())
    })

    // periodically log AgentManager state
    setInterval(() => {
      Logger.info(`Monitoring ${this.list().length} connected NSolid Agents.`)
    }, LOG_INTERVAL).unref()

    this._agents.on('added', (id, agent) => {
      const app = agent.info ? agent.info.app : '... n/a yet ...'
      Logger.debug(`added agent ${id}: app: ${app}`)

      // Agents that start up quickly and send metrics may not yet be
      // subscribed completely, so delaying this request is necessary
      setTimeout(() => {
        // No actual callback required, response is handled by zmq handler.
        if (this.byId(agent.id)) {
          sendCB({
            command: 'packages',
            id: agent.id
          })
        }
      }, 1000).unref()

      if (agent.info == null) {
        Logger.debug('no info for %s', agent.id)
        // Agents that start up quickly and send metrics may not yet be
        // subscribed completely, so delaying this request is necessary
        setTimeout(() => {
          // No actual callback required, response is handled by zmq handler.
          if (this.byId(agent.id)) {
            sendCB({
              command: 'info',
              id: agent.id
            })
          }
        }, 1000).unref()
      }
    })

    this._agents.on('expired', (id, agent) => {
      const eventArgs = Agent.getEventArgs(agent)
      this._context.eventManager.emitSystemEvent('agent-missing', agent.id, eventArgs)
      Logger.debug(`agent ${id} is now considered "missing"`)
      this.addLifeToTotal(agent)
    })

    this._agents.on('deleted', (id, agent) => {
      this.addLifeToTotal(agent)
    })

    const self = this

    // commandDispatcher can be null in tests
    function sendCB (message, timeout, cb) {
      if (self._context.commandDispatcher == null) return

      // or set to an object without the `sendCB()` method
      if (self._context.commandDispatcher.sendCB == null) return

      self._context.commandDispatcher.sendCB(message, timeout, cb)
    }
  }

  // update the vulns from the cache for all the agents
  updateVulns () {
    for (let agent of this.list()) {
      agent.updateVulnsFromCache()
    }

    this._context.vulnManager.updateActiveVulns()
  }

  addLifeToTotal (agent) {
    this._totalAgentLifetime += agent.aliveMs
    this._totalAgentActiveMs += agent.busyMs
  }

  // Return the agent with specified id.
  byId (id) {
    return this._agents.get(id)
  }

  deadById (id) {
    return this._deadAgents.get(id)
  }

  // Return the agent with specified id.
  byVulnId (vulnId) {
    return this.list().filter(agent => {
      const vulns = agent.vulns.filter(vuln => vuln.id === vulnId)
      return vulns.length !== 0
    })
  }

  byPackages (hash) {
    if (!hash) {
      return []
    }

    return this.list().filter(agent => {
      return agent.packagesHash && agent.packagesHash === hash
    })
  }

  // Add an agent.
  upsert (agent) {
    Logger.debug(`adding agent ${agent.id}`)
    this._agents.upsert(agent.id, agent)
  }

  // Remove an agent.
  remove (agent) {
    Logger.debug(`removing agent ${agent.id}`)
    agent.dispose()
    this._agents.delete(agent.id)

    this._context.vulnManager.updateActiveVulns()
  }

  markDead (agent) {
    this.remove(agent)
    this._deadAgents.upsert(agent.id, agent)
  }

  // Return an array of all the agents.
  list () {
    const agents = this._agents.entries()
    return Object.keys(agents).map((key) => {
      return agents[key]
    })
  }

  count () {
    return this._agents.count()
  }

  // for tests
  reset () {
    this._agents.purge()
    this._deadAgents.purge()
  }

  // get the info record for an agent from cache or influx
  getInfoById (id, cb) {
    // if the agent is cached, return that
    const cachedAgent = this.byId(id)
    this._context.settingsManager.getEntry('agents', id, (err, cachedInfo) => {
      if (err) return (cb, err)
      if (cachedAgent || cachedInfo) {
        let info = {
          time: Date.now()
        }
        info = Object.assign(info, cachedAgent ? cachedAgent.info : cachedInfo)
        cb(null, info)
        return
      }

      err = new Error(`No cached agent info for ${id}`)
      Logger.warn(err)
      cb(err)
    })
  }

  query (queryExpr) {
    const agents = this._agents.entries()
    return Object.keys(agents).reduce((list, key) => {
      const agent = agents[key]
      if (agent.metrics == null || agent.info == null) return list
      if (queryIsEmpty(queryExpr) || queryExpr.evalForAgent(agent)) {
        list.push(agent)
      }
      return list
    }, [])
  }
}

function queryIsEmpty (queryExpr) {
  return queryExpr == null || !queryExpr.terms || queryExpr.terms.length === 0
}
