'use strict'

const MillisecondsPerHour = 60 * 60 * 1000

class Report {
  constructor (agentManager) {
    this.agentManager = agentManager
  }

  agentHours () {
    // if this is going to be reliable, it needs to be idempotently computed
    // so we should count the total runtime of agents that have come and gone
    // plus the computed uptime of agents that are still in cache
    const agents = _values(this.agentManager._agents.entries())
    const totalCurrentUptime = this.agentManager._totalAgentLifetime +
      agents.reduce((sum, agent) => {
        return sum + (agent.aliveMs || 0)
      }, 0)

    return totalCurrentUptime / MillisecondsPerHour
  }

  agentActiveHours () {
    const agents = _values(this.agentManager._agents.entries())
    const totalActiveTime = this.agentManager._totalAgentActiveMs +
      agents.reduce((sum, agent) => {
        return sum + (agent.busyMs || 0)
      }, 0)

    return totalActiveTime / MillisecondsPerHour
  }

  medianUptime () {
    const agents = _values(this.agentManager._agents.entries())
    const uptimes = agents.map(agent => { return agent.metrics ? agent.metrics.uptime : 0 })
    if (uptimes.length === 0) return 0
    const sorted = uptimes.sort((a, b) => { return b - a })
    const median = sorted[Math.floor(sorted.length / 2)]
    return median
  }

  numHosts () {
    const hosts = new Set()
    const agents = _values(this.agentManager._agents.entries())
    for (let agent of agents) {
      if (agent.info && agent.info.hostname) {
        hosts.add(agent.info.hostname)
      }
    }
    return hosts.size
  }

  numCores () {
    const seenHosts = new Set()
    const agents = _values(this.agentManager._agents.entries())
    return agents.reduce((sum, agent) => {
      if (!(agent.info && agent.info.cpuCores)) {
        return sum
      }
      if (seenHosts.has(agent.info.hostname)) {
        return sum
      }
      seenHosts.add(agent.info.hostname)
      return sum + agent.info.cpuCores
    }, 0)
  }

  numApplications () {
    const agents = _values(this.agentManager._agents.entries())
    const apps = new Set(
      agents.map(agent => {
        return (agent.info && agent.info.app) || 'unknown'
      })
    )
    return apps.size
  }

  versions () {
    const versions = {}
    const agents = _values(this.agentManager._agents.entries())
    for (let agent of agents) {
      if (!(agent.info && agent.info.versions)) return
      const version = agent.info.versions.nsolid
      if (!versions[version]) {
        versions[version] = 0
      }
      versions[version]++
    }
    return versions
  }
}

exports.Report = Report

function _values (object) {
  return Object.keys(object).map(k => object[k])
}
