'use strict'

const lodash = require('lodash')

const RevocableEventEmitter = require('./RevocableEventEmitter')

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

exports.create = create
exports.calculatePHMU = calculatePHMU
exports.calculatePHMUByProcess = calculatePHMUByProcess
exports.getEventArgs = getEventArgs

// Create and return a new N|Solid agent.
function create (context, id, info, metrics) {
  return new Agent(context, id, info, metrics)
}

// Models an N|Solid agent.
class Agent extends RevocableEventEmitter {
  constructor (context, id, info, metrics) {
    super()

    this._context = context
    this.id = id
    this.tags = {}
    this.info = info // may be null
    this.metrics = metrics // may be null
    this._queuedMetrics = []
    this.packagesHash = null
    this.packages = []
    this.vulns = []
    this.pendingCommandTimers = new Set()
    this.busyMs = 0

    // This will simply add the number of reporting periods that this agent
    // cache entry collected.  Using agent uptime or lifetime has issues with
    // clock skew, agent cache expiry and reconnect, and more.  Further, by
    // only incrementing when metrics are accrued, we should only be counting a
    // process as alive when we know that it was.
    this.aliveMs = 0

    if (info) {
      this.addInfo(info)
    }

    if (metrics) {
      this.addMetrics(metrics)
    }

    Logger.debug(`creating new agent: ${id}`)
  }

  // update the vuln info from cache, after it's been updated
  updateVulnsFromCache () {
    const packageInfo = this._context.packageManager.getPackageInfo(this.packagesHash)

    if (packageInfo == null) {
      Logger.warn(`unable to find package info during updatePackageInfo() for ${this.id}`)
      return
    }

    this.vulns = packageInfo.vulns
  }

  // called when packages message received from agent
  addPackages (packages) {
    if (!packages || !packages.length) {
      this.packages = []
      this.vulns = []
      this.packagesHash = this._context.packageManager.hashPackages(this.packages)
      this.emit('packages-added', this)
      this._context.eventManager.emitSystemEvent('agent-packages-added', this.id, getEventArgs(this))
      return
    }

    this.packagesHash = this._context.packageManager.hashPackages(packages)
    const packageInfo = this._context.packageManager.getPackageInfo(this.packagesHash)

    // shouldn't happen
    if (packageInfo == null) return

    this.packages = packageInfo.packages
    this.vulns = packageInfo.vulns
    this.emit('packages-added', this)
    this._context.eventManager.emitSystemEvent('agent-packages-added', this.id, getEventArgs(this))

    const vulnIds = this.vulns.map(vuln => vuln.id)
    this._context.fieldManager.updateRanges('packages', { vuln: vulnIds })
    this._context.vulnManager.updateActiveVulns()
  }

  addInfo (message) {
    if (!message || !message.body) {
      return
    }

    if (!this.info) {
      setImmediate(() => {
        const eventArgs = getEventArgs(this)
        this._context.eventManager.emitSystemEvent('agent-found', this.id, eventArgs)
      })
    }

    this.info = lodash.clone(message.body)
    this._context.fieldManager.updateRanges('info', this.info)

    if (message.body.id) {
      this.tags.id = message.body.id
    }

    if (message.body.tags) {
      this.tags.tags = `,${message.body.tags.join(',')},`
    }

    if (message.body.app) {
      this.tags.app = message.body.app
      delete message.body.app
    }

    if (message.body.hostname) {
      this.tags.hostname = message.body.hostname
      delete message.body.hostname
    }

    const item = formatInfoForDb(this, message)
    this._context.settingsManager.setSetting('agents', item, err => {
      if (err) Logger.error('Could not save agent info', err, message)
    })

    if (this._queuedMetrics.length) {
      this._queuedMetrics.forEach(metrics => {
        metrics.body = this.formatMetricsForWriting(metrics.body)
        this._context.influxdbWriter.writeMetrics(this, metrics)
      })
      this._queuedMetrics = []
    }
  }

  addMetrics (message) {
    if (!message || !message.body) {
      return
    }

    const interval = message.interval || 3000

    this.aliveMs += interval

    if (message.body.hasOwnProperty('loopIdlePercent')) {
      const busyMs = interval * ((100 - message.body.loopIdlePercent) / 100)
      this.busyMs += busyMs
    }

    message.body.vulns = this.vulns.length

    if (this.info && this._context.influxdbWriter) {
      this.metrics = message.body = this.formatMetricsForWriting(message.body)
      this._context.fieldManager.updateRanges('metrics', this.metrics)

      this._context.influxdbWriter.writeMetrics(this, message)
    } else {
      this._queuedMetrics.push(message)
    }
  }

  // Used when writing metrics to the time-series database.
  formatMetricsForWriting (metrics) {
    // convert cpu % from 0..100 to 0..1 for consistency with other
    // percentile values
    if (metrics.cpu != null) {
      metrics.cpu = metrics.cpu / 100
    }

    return metrics
  }

  // Used when returning metrics from POST /metrics command.
  formatMetricsForReturning (metrics) {
    metrics = this.formatMetricsForWriting(metrics)
    metrics = this.addPercentMemoryProperties(metrics)

    if (this.info != null) {
      metrics.app = this.info.app
      metrics.hostname = this.info.hostname
      metrics.tags = this.info.tags
    }

    return metrics
  }

  // Used when returning metrics from GET /metrics command.
  formatMetricsFromQuery (metrics) {
    metrics = this.addPercentMemoryProperties(metrics)

    if (this.info != null) {
      metrics.app = this.info.app
      metrics.hostname = this.info.hostname
      metrics.tags = this.info.tags
    }

    return metrics
  }

  // Adds the calculated fields to metrics using info.
  addPercentMemoryProperties (metrics) {
    if (!metrics || !this.info) return metrics

    if (metrics.rss != null) {
      metrics.percentHostMemoryUsedByProcess = calculatePHMUByProcess({
        rss: metrics.rss,
        totalMem: this.info.totalMem
      })
    }

    if (metrics.freeMem != null) {
      metrics.percentHostMemoryUsed = calculatePHMU({
        totalMem: this.info.totalMem,
        freeMem: metrics.freeMem
      })
    }

    return metrics
  }

  getCacheRecord () {
    var obj = {
      id: this.id,
      app: this.tags.app,
      hostname: this.tags.hostname,
      tags: canonicalizeTags(this.info && this.info.tags),
      info: this.info,
      metrics: this.metrics
    }
    return obj
  }

  dispose () {
    for (var timer of this.pendingCommandTimers) {
      clearTimeout(timer)
      this.pendingCommandTimers.delete(timer)
    }
  }
}

function calculatePHMUByProcess (params) {
  if (!params.rss || !params.totalMem) {
    return 0
  }
  return Math.round(10000 * params.rss / params.totalMem) / 10000
}

function calculatePHMU (params) {
  if (!params.freeMem || !params.totalMem) {
    return 0
  }
  return Math.round(10000 * (params.totalMem - params.freeMem) / params.totalMem) / 10000
}

// Normalize tags as null, an array, or string, to an array.
function canonicalizeTags (tags) {
  if (tags == null) return []
  if (Array.isArray(tags)) return tags

  // sometimes tags have leading and trailing commas ...
  return tags.split(/,/g).filter((tag) => tag !== '')
}

function formatInfoForDb (agent, message) {
  // Ensure sane shape of body before modifying it
  const data = lodash.defaults({}, message.body)
  data.id = agent.id
  data.app = agent.info.app
  data.tags = agent.info.tags
  data.hostname = agent.info.hostname

  return data
}

function getEventArgs (agent) {
  const eventArgs = {}

  if (agent.info) {
    if (agent.info.app) eventArgs.app = agent.info.app
    if (agent.info.hostname) eventArgs.hostname = agent.info.hostname
    if (agent.info.tags) eventArgs.tags = agent.info.tags
    if (agent.info.pid) eventArgs.pid = agent.info.pid
  }

  return eventArgs
}
