'use strict'

const EventEmitter = require('events')
const lodash = require('lodash')
const pg = require('pg')
const url = require('url')

exports.create = create

const types = ['integrations', 'generalSettings', 'notifications', 'savedViews']

const integrationsCols = 'licensekey, id, type, name, url, key, emails, created, modified'
const integrationsReturnCols = 'id, type, name, url, key, emails'
// excluded.license is omitted so the licensekey can never change
const integrationsSetQuery = `insert into integrations (${integrationsCols})
  values ($1, $2, $3, $4, $5, $6, $7, NOW(), NOW())
  on conflict (id) do update
  set type = excluded.type, name = excluded.name, url = excluded.url,
    key = excluded.key, emails = excluded.emails,
    created = excluded.created, modified = excluded.modified
  returning ${integrationsReturnCols}`
const integrationsDeleteQuery = `delete from integrations where
  licensekey = $1 and id = $2`
const integrationsGetQuery = `select * from integrations where
  licensekey = $1 and id = $2 limit 1`
const integrationsListQuery = `select * from integrations where
  licensekey = $1`

const generalSettingsCols = 'licensekey, id, value, created, modified'
const generalSettingsReturnCols = 'id, value'
// excluded.license is omitted so the licensekey can never change
const generalSettingsSetQuery = `insert into generalsettings (${generalSettingsCols})
  values ($1, $2, $3, NOW(), NOW())
  on conflict (licensekey, id) do update
  set value = excluded.value,
    created = excluded.created, modified = excluded.modified
  returning ${generalSettingsReturnCols}`
const generalSettingsDeleteQuery = `delete from generalsettings where
  licensekey = $1 and id = $2`
const generalSettingsGetQuery = `select * from generalsettings where
  licensekey = $1 and id = $2 limit 1`
const generalSettingsListQuery = `select * from generalsettings where
  licensekey = $1`

const notificationsCols = 'licensekey, event, actions, threshold, onlyexceptions, exittype, created, modified'
const notificationsReturnCols = 'event, actions, threshold, onlyexceptions, exittype'
// excluded.license is omitted so the licensekey can never change
const notificationsSetQuery = `insert into notifications (${notificationsCols})
  values ($1, $2, $3, $4, $5, $6, NOW(), NOW())
  on conflict (licensekey, event) do update
  set actions = excluded.actions, threshold = excluded.threshold,
    onlyexceptions = excluded.onlyexceptions, exittype = excluded.exittype,
    created = excluded.created, modified = excluded.modified
  returning ${notificationsReturnCols}`
const notificationsDeleteQuery = `delete from notifications where
  licensekey = $1 and event = $2`
const notificationsGetQuery = `select * from notifications where
  licensekey = $1 and event = $2 limit 1`
const notificationsListQuery = `select * from notifications where
  licensekey = $1`

const savedViewsCols = 'licensekey, id, name, expression, axisx, axisy, snooze, delay, actions, persistent, created, modified'
const savedViewsReturnCols = 'id, name, expression, axisx, axisy, snooze, delay, actions, persistent'
// excluded.license is omitted so the licensekey can never change
const savedViewsSetQuery = `insert into savedviews (${savedViewsCols})
  values ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, NOW(), NOW())
  on conflict (id) do update
  set name = excluded.name, expression = excluded.expression,
    axisx = excluded.axisx, axisy = excluded.axisy, snooze = excluded.snooze,
    delay = excluded.delay, actions = excluded.actions,
    persistent = excluded.persistent,
    created = excluded.created, modified = excluded.modified
  returning ${savedViewsReturnCols}`
const savedViewsDeleteQuery = `delete from savedviews where
  licensekey = $1 and id = $2`
const savedViewsGetQuery = `select * from savedviews where
  licensekey = $1 and id = $2 limit 1`
const savedViewsListQuery = `select * from savedviews where
  licensekey = $1`

function create (context) {
  return new PostGresConnector(context)
}

function dbDsn (config) {
  const params = url.parse(config.data.postgresDsn)
  let newConfig = {
    idleTimeoutMillis: 1000,
    host: params.hostname,
    port: params.port,
    database: params.pathname.split('/')[1],
    ssl: !!config.data.postgresSsl // for local docker dev
  }
  const useAuth = !!params.auth
  const auth = useAuth ? params.auth.split(':') : []
  if (useAuth && auth.length === 2) {
    newConfig = Object.assign({}, newConfig, {
      user: auth[0],
      password: auth[1]
    })
  }
  return newConfig
}

// Manages the application settings.
class PostGresConnector extends EventEmitter {
  constructor (context) {
    super()

    this._context = context
    this._pool = new pg.Pool(dbDsn(context.config))

    this._pool.on('connect', _ => {
      this.emit('connect')
    })

    this.setupDbApi()
  }

  getLicenseKey () {
    return lodash.get(this, '_context.provider.license.key')
  }

  extractRows (fn) {
    return function rowExtractor (err, result) {
      if (err) {
        return fn(err)
      }
      fn(null, (result && result.rows) || [])
    }
  }

  // Returns the agents stored in leveldb, in the callback.
  // The callback is invoked for every agent returned from leveldb.
  // On error, the callback is invoked with the error from leveldb, and
  // no more agents will be returned.
  // When there are no more agents to be returned, a null will be passed
  // to the callback instead of an agent.
  streamAgents (cb) {
    return setImmediate(cb, new Error('Implement me!'))
  }

  setupDbApi () {
    this.delete = {}
    this.delete.integrations = this.deleteIntegration.bind(this)
    this.delete.generalSettings = this.deleteGeneralSettings.bind(this)
    this.delete.notifications = this.deleteNotification.bind(this)
    this.delete.savedViews = this.deleteSavedView.bind(this)

    this.set = {}
    this.set.integrations = this.setIntegration.bind(this)
    this.set.generalSettings = this.setGeneralSettings.bind(this)
    this.set.notifications = this.setNotification.bind(this)
    this.set.savedViews = this.setSavedView.bind(this)

    this.get = {}
    this.get.integrations = this.getIntegration.bind(this)
    this.get.generalSettings = this.getGeneralSettings.bind(this)
    this.get.notifications = this.getNotification.bind(this)
    this.get.savedViews = this.getSavedView.bind(this)

    this.list = {}
    this.list.integrations = this.listIntegrations.bind(this)
    this.list.generalSettings = this.listGeneralSettings.bind(this)
    this.list.notifications = this.listNotifications.bind(this)
    this.list.savedViews = this.listSavedViews.bind(this)
  }

  getSettings (type, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (type === 'license') { // TODO remove me
      return setImmediate(cb, null, { key: this.getLicenseKey() })
    } else if (type === 'agents') { // TODO remove me
      return setImmediate(cb, null, [])
    } else if (types.indexOf(type) === -1) {
      return setImmediate(cb, new Error(`Invalid value ${type} for required argument 'type'.`))
    }

    this.list[type]((err, res) => {
      if (err) return cb(err)
      if (!res) return cb(null, {})

      const obj = {}
      res.forEach(item => {
        obj[getId(type, item)] = item
      })
      cb(null, obj)
    })
  }

  // Get a single entry by ID
  getEntry (type, id, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (!id) {
      return setImmediate(cb, new Error('Missing required argument `id`.'))
    } else if (types.indexOf(type) === -1) {
      return setImmediate(cb, new Error('Invalid value for required argument `type`.'))
    }

    return this.get[type](id, cb)
  }

  setSetting (type, item, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (!item) {
      return setImmediate(cb, new Error('Missing required argument `item`.'))
    } else if (types.indexOf(type) === -1) {
      return setImmediate(cb, new Error(`Invalid value ${type} for required argument 'type'.`))
    }

    this.set[type](item, (err, items) => {
      if (err) return cb(err)
      cb(null, items[0])
    })
  }

  deleteSetting (type, id, cb) {
    if (!type) {
      return setImmediate(cb, new Error('Missing required argument `type`.'))
    } else if (!id) {
      return setImmediate(cb, new Error('Missing required argument `id`.'))
    } else if (types.indexOf(type) === -1) {
      return setImmediate(cb, new Error(`Invalid value ${type} for required argument 'type'.`))
    }

    if (type === 'savedViews') {
      const cached = this.get[type](this._settings, type, id)
      if (cached && cached.persistent === true) {
        return setImmediate(cb, new Error('Can not remove default saved view.'))
      }
    }

    this.delete[type](id, cb)
  }

  runQuery (query, params, cb) {
    if (query.indexOf('licensekey') === -1) {
      return setImmediate(cb, new Error('Database query must include license key.'))
    }

    this._pool.connect((err, client, done) => {
      if (err) {
        done()
        return setImmediate(cb, err)
      }

      client.query(query, params, this.extractRows((err, res) => {
        done()
        setImmediate(cb, err, res)
      }))
    })
  }

  setIntegration (item, cb) {
    const params = [this.getLicenseKey(), item.id, item.type, item.name, item.url, item.key, item.emails]
    this.runQuery(integrationsSetQuery, params, cb)
  }

  setGeneralSettings (item, cb) {
    const params = [this.getLicenseKey(), item.id, item.value]
    this.runQuery(generalSettingsSetQuery, params, cb)
  }

  setNotification (item, cb) {
    const params = [this.getLicenseKey(), item.event, JSON.stringify(item.actions), item.threshold, item.onlyexceptions, item.exittype]
    this.runQuery(notificationsSetQuery, params, cb)
  }

  setSavedView (item, cb) {
    const params = [this.getLicenseKey(), item.id, item.name, item.expression, item.axisx, item.axisy, item.snooze, item.delay, JSON.stringify(item.actions), item.persistent]
    this.runQuery(savedViewsSetQuery, params, cb)
  }

  deleteIntegration (id, cb) {
    const params = [this.getLicenseKey(), id]
    this.runQuery(integrationsDeleteQuery, params, cb)
  }

  deleteGeneralSettings (id, cb) {
    const params = [this.getLicenseKey(), id]
    this.runQuery(generalSettingsDeleteQuery, params, cb)
  }

  deleteNotification (id, cb) {
    const params = [this.getLicenseKey(), id]
    this.runQuery(notificationsDeleteQuery, params, cb)
  }

  deleteSavedView (id, cb) {
    const params = [this.getLicenseKey(), id]
    this.runQuery(savedViewsDeleteQuery, params, cb)
  }

  getIntegration (item, cb) {
    const params = [this.getLicenseKey(), item.id]
    this.runQuery(integrationsGetQuery, params, cb)
  }

  getGeneralSettings (item, cb) {
    const params = [this.getLicenseKey(), item.id]
    this.runQuery(generalSettingsGetQuery, params, cb)
  }

  getNotification (item, cb) {
    const params = [this.getLicenseKey(), item.event]
    this.runQuery(notificationsGetQuery, params, cb)
  }

  getSavedView (item, cb) {
    const params = [this.getLicenseKey(), item.id]
    this.runQuery(savedViewsGetQuery, params, cb)
  }

  listIntegrations (cb) {
    const params = [this.getLicenseKey()]
    this.runQuery(integrationsListQuery, params, cb)
  }

  listGeneralSettings (cb) {
    const params = [this.getLicenseKey()]
    this.runQuery(generalSettingsListQuery, params, cb)
  }

  listNotifications (cb) {
    const params = [this.getLicenseKey()]
    this.runQuery(notificationsListQuery, params, cb)
  }

  listSavedViews (cb) {
    const params = [this.getLicenseKey()]
    this.runQuery(savedViewsListQuery, params, cb)
  }
}

/*
 * Different types of data use a different attribute for their unique key.
 */
function getId (type, item) {
  let id
  switch (type) {
    case 'notifications':
      id = item.event
      break
    case 'integrations':
    case 'savedViews':
    case 'generalSettings':
    default:
      id = item.id
      break
  }

  if (!id) throw new Error('Invalid settings format.')
}
