'use strict'

exports.getDepsChainsMap = getDepsChainsMap

const path = require('path')

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

// given an agent's packages property, and it's vulns, return a Map
// vulnId -> [ depsChain ], where a depsChain is an array of package
// which is an array of package@version strings of package roots to the
// vulnerable package.
function getDepsChainsMap (packages, vulns) {
  const result = new Map()

  if (packages == null || vulns == null) return result
  if (packages.length === 0 || vulns == null) return result

  // convert flat packages array to nice map of objects
  const packagesMap = reifyPackages(packages)

  // get array of package objects which are the "roots"
  const packageRoots = getPackageRoots(packagesMap)
  Logger.debug(`packageRoots: ${packageRoots.map(pkg => pkg.path).join(', ')}`)

  for (let vuln of vulns) {
    const depsChains = getDepsChains(packagesMap, packageRoots, vuln.id)
    result.set(vuln.id, depsChains)
  }

  return result
}

function getDepsChains (packagesMap, packageRoots, vulnId) {
  // is a package root the vulnerable module, if so, got our chain
  for (let packageRoot of packageRoots) {
    if (packageHasVuln(packageRoot, vulnId)) {
      return [ [ packageNameVersion(packageRoot) ] ]
    }
  }

  const result = []

  // build initial list of depsChains which is each root and it's top-level deps
  const initialChains = []
  for (let packageRoot of packageRoots) {
    for (let dep of packageRoot.dependencies) {
      // if the dep is vulnerable, just set it, otherwise push it for processing
      if (packageHasVuln(dep, vulnId)) {
        result.push([packageNameVersion(packageRoot), packageNameVersion(dep)])
      } else {
        initialChains.push([packageRoot, dep])
      }
    }
  }

  // for each initialChain, resolve to the shortest chain that includes the vuln
  for (let initialChain of initialChains) {
    const seen = new Set(initialChain)
    const shortestChain = getShortestChain(initialChain, vulnId, seen)
    if (shortestChain == null) continue

    // convert array of pkg objects to name@version
    result.push(shortestChain.map(packageNameVersion))
  }

  return result
}

// Find the shorest chain of all chains from passed in chain to package with vuln.
// Returns null if there is no path at all.
// chain is an array of package objects
function getShortestChain (chain, vulnId, seen) {
  const allChains = []
  getAllChains(chain, vulnId, seen, allChains)

  // if no chains, return null
  if (allChains.length == null) return null

  // sort the chains by length
  allChains.sort((a, b) => a.length - b.length)

  // return the shortest
  return allChains[0]
}

function getAllChains (chain, vulnId, seen, allChains) {
  const currPkg = chain[chain.length - 1]

  for (let dep of currPkg.dependencies) {
    const depChain = chain.concat(dep)
    if (packageHasVuln(dep, vulnId)) {
      allChains.push(depChain)
      return
    }

    if (seen.has(dep)) continue

    seen.add(dep)
    getAllChains(depChain, vulnId, seen, allChains)
  }
}

// return pkgName@version for a package
function packageNameVersion (pkg) {
  return `${pkg.name}@${pkg.version}`
}

// return boolean of whether package has specified vuln
function packageHasVuln (pkg, vulnId) {
  return pkg.vulns.indexOf(vulnId) !== -1
}

// return array of package objects which are the "roots" of the packages graph
function getPackageRoots (packagesMap) {
  // create array of [dependents.length, package]
  const depsPkgArray = Array.from(packagesMap.values())
    .map(pkg => {
      const dependents = pkg.dependents || []
      return [dependents.length, pkg]
    })
    .sort((a, b) => a[0] - b[0])

  // if no packages, return empty array
  if (depsPkgArray.length === 0) return []

  // return all the entries with the smallest number of deps
  const minDeps = depsPkgArray[0][0]

  const result = []
  for (let depsPkg of depsPkgArray) {
    if (depsPkg[0] !== minDeps) break
    result.push(depsPkg[1])
  }

  return result
}

// convert the array of "flat" package objects to a map of pkg.path -> package,
// where the package object has references to other package objects for the
// deps isntead of relative path names.
function reifyPackages (packages) {
  // build the initial map with a shallow copy of the package
  const packagesMap = new Map()
  for (let pkg of packages) {
    packagesMap.set(pkg.path, Object.assign({}, pkg))
  }

  // convert deps from relative package names to package objects
  for (let pkg of packagesMap.values()) {
    pkg.dependencies = pkg.dependencies
      .map(relPath => getPackage(pkg.path, relPath))
      .filter(pkg => pkg != null)  // handle bad data, shouldn't happen!
  }

  return packagesMap

  function getPackage (packagePath, relPath) {
    const absPath = path.resolve(packagePath, relPath)
    return packagesMap.get(absPath)
  }
}
