'use strict'

// Sends a file as an HTTP response
// If file isn't gzipped, just send down un-gzipped version.
// If file is gzipped, but request doesn't support it, need to pipe the file
// through an gunzip filter.

exports.sendFile = sendFile

const fs = require('fs')
const zlib = require('zlib')

const util = require('../util')
const Logger = util.getLogger(__dirname, __filename)

// send a gzipped file as an express response
function sendFile (req, res, fileName, isCompressed, headers, cb) {
  if (cb == null) {
    cb = function () {}
  } else {
    cb = util.onlyCallOnce(cb)
  }

  // if file is not gzipped, send it ungzipped, even if gzip accepted
  if (!isCompressed) {
    Logger.debug(`sending ungzipped file ${fileName}`)
    return sendFileRes(res, fileName, headers, cb)
  }

  // does the client accept gzip?
  const acceptsGzip = req.acceptsEncodings('gzip')

  // if so, just send the file, with content-encoding header
  if (acceptsGzip) {
    Logger.debug(`sending gzipped file ${fileName}`)
    headers['Content-Encoding'] = 'gzip'
    return sendFileRes(res, fileName, headers, cb)
  }

  // client did not request gzip, so need to ungzip
  Logger.debug(`piping gzipped file: ${fileName}`)

  // first check file
  fs.access(fileName, (fs.constants || fs).R_OK, gotAccess)

  // pipe the unzipped file
  function gotAccess (err) {
    if (err) {
      Logger.error(err, `requested asset ${fileName} not readable`)
      sendErrorResponse(res)
      return cb(err)
    }

    // create unzip stream
    const unzipStream = zlib.createGunzip()
    unzipStream.on('error', err => {
      Logger.error(err, `error unzipping ${fileName}`)
      sendErrorResponse(res)
      return cb(err)
    })

    // create the file stream
    const fileStream = fs.createReadStream(fileName)
    fileStream.on('error', err => {
      Logger.error(err, `error reading file ${fileName}`)
      sendErrorResponse(res)
      return cb(err)
    })

    // send headers, pipe streams as response
    res.set(headers)
    res.on('finish', () => cb())

    fileStream.pipe(unzipStream).pipe(res)
  }
}

// send a file as a response
function sendFileRes (res, fileName, headers, cb) {
  res.sendFile(fileName, { headers }, (err) => {
    if (!err) return cb()

    // log message if it wasn't user closing the socket
    if (['EPIPE', 'ECONNABORTED'].indexOf(err.code) === -1) {
      Logger.warn(err, 'error sending file')
    }

    sendErrorResponse(res)
    cb(err)
  })
}

// send an error response, if headers not sent, and end response
function sendErrorResponse (res) {
  if (res.headersSent) {
    try {
      res.end()
    } catch (err) {
      Logger.warn(err, 'error ending response')
    }
    return
  }

  try {
    res.status(404).send('asset not found')
  } catch (err) {
    Logger.warn(err, `error sending not found response`)
  }
}
