'use strict'

exports.handleRequest = handleRequest

const normalizers = require('./normalizers')

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context
  const eventStream = req.nsConsole.eventStream
  const qData = req.nsConsole.qData

  Logger.debug('handling request')

  // default series is 'raw'
  if (qData.series == null) qData.series = 'raw'

  // set up the date start/end values
  if (qData.start == null) {
    switch (qData.series) {
      case 'raw':
        // 1 minute, about 20 records per agent
        qData.start = Date.now() - 60 * 1000
        break
      case '1m':
        // 20 minutes, about 20 records per agent
        qData.start = Date.now() - 20 * 60 * 1000
        break
      case '1h':
        // 20 hours, about 20 records per agent
        qData.start = Date.now() - 20 * 60 * 60 * 1000
        break
      default:
        // unexpected! treat like 'raw'
        qData.start = Date.now() - 60 * 1000
    }
  }
  if (qData.end == null) qData.end = Date.now()

  qData.start = util.dateMS2ISO(qData.start)
  qData.end = util.dateMS2ISO(qData.end)

  if (qData.start > qData.end) {
    eventStream.writeError(null, undefined, `start date ${qData.start} greater than end date ${qData.start}`)
    eventStream.close()
    return
  }

  if (context.influxdbReader == null) {
    eventStream.writeError(null, undefined, 'influx is not available')
    eventStream.close()
    return
  }

  const query = buildQuery(context, qData)
  Logger.debug(`query: ${query}`)
  context.influxdbReader.readMetrics(query, readMetricsDone)

  // handle result of influx query
  function readMetricsDone (err, data) {
    if (err) {
      Logger.error(err, 'error reading metrics')
      eventStream.writeError(null, undefined, 'error reading metrics')
      eventStream.close()
      return
    }

    if (eventStream.isClosed()) return

    const records = recordsFromQuery(data)
    for (let record of records) {
      const metrics = influxRecordToMetrics(record)
      eventStream.writeData(metrics)
    }

    eventStream.close()
  }
}

// convert influx record to metrics record
function influxRecordToMetrics (record) {
  const metrics = normalizers.agentInfo(record)
  return metrics
}

// create metrics records from the query results.
function recordsFromQuery (data) {
  try {
    data = JSON.parse(data)
  } catch (err) {
    Logger.warn(err, 'error parsing JSON data from query', data)
    return []
  }

  Logger.debug(`query results: ${JSON.stringify(data, null, 4)}`)

  if (data.results == null) return []
  if (data.results.length === 0) return []
  if (data.results[0].series == null) return []
  if (data.results[0].series.length === 0) return []

  const series = data.results[0].series[0]
  const records = []

  Logger.debug(`recordsFromQuery(): received ${series.values.length} values`)
  for (let i = 0; i < series.values.length; i++) {
    var record = {}
    for (let j = 0; j < series.columns.length; j++) {
      if (series.values[i][j] == null) continue
      record[series.columns[j]] = series.values[i][j]
    }

    records.push(record)
  }

  // TODO: fix tags and perhaps CPU
  return records
}

// build the influx query
function buildQuery (context, qData) {
  const queryExpr = qData.q
  const queryParts = ['SELECT']

  // currently returning all fields; code like below can be used to select
  // exact fields to be returned

  // const fixedFields = 'id app hostname tags'.split(' ')

  // if (qData.fields == null || qData.fields.length === 0) qData.fields = []
  // for (let fixedField of fixedFields) {
  //   if (qData.fields.indexOf(fixedField) === -1) qData.fields.push(fixedField)
  // }

  // if (qData.fields.length === 0) {
  //   queryParts.push('*')
  // } else {
  //   const fieldSpec = qData.fields
  //     .map(field => `"${field}"`)
  //     .join(',')
  //   queryParts.push(fieldSpec)
  // }

  let series = 'agentData.metrics'
  if (qData.series === '1m') {
    series = 'one_minute_daily.metrics'
  }
  if (qData.series === '1h') {
    series = 'one_hour_monthly.metrics'
  }

  queryParts.push('*')
  queryParts.push(`FROM ${series} WHERE`)

  const whereParts = []

  // needs to check all query terms, for now just checking id={agentId}
  if (queryExpr && queryExpr.terms) {
    for (let filterTerm of queryExpr.terms) {
      if (filterTerm.field.name === 'id') {
        // Normalize id.  To use a short form, the agent must still be running and
        // in cache. If we end up searching for a short id anyway, use a regex
        let id, op
        const agent = context.agentManager.byId(filterTerm.values[0])
        if (agent) {
          id = `'${agent.id}'`
          op = '='
        } else {
          id = `/^${filterTerm.values[0]}/`
          op = '=~'
        }

        whereParts.push(`id ${op} ${id}`)
      }
    }
  }

  whereParts.push(`time >= '${qData.start}'`)
  whereParts.push(`time <= '${qData.end}'`)

  influxWhereTerms(qData.filterExpression).forEach(term => {
    whereParts.push(term)
  })

  queryParts.push(whereParts.join(' AND '))

  return queryParts.join(' ')
}

// return an array of ANDable influx expressions for the filter expression
// TODO: complete - note that tags should be regex expression
function influxWhereTerms (filterExpr) {
  return []
}
