'use strict'

const path = require('path')
const http = require('http')
const https = require('https')
const QueryString = require('querystring')

const engineIO = require('engine.io')
const lodash = require('lodash')
const express = require('express')
const bodyParser = require('body-parser')

const util = require('./util')
const qsp = require('./http-api/v3/query-string-parser')
const EventStream = require('./http-api/event-stream')
const fileSender = require('./http-api/file-sender')

exports.launch = launch

const Logger = util.getLogger(__dirname, __filename)
const LOG_INTERVAL = 30 * 1000 // 30 seconds
const runningQueries = new Set()
var totalQueries = 0

const API_VERSIONS = ['3']

// Launch an HTTP(S) server.
function launch (context, cb) {
  cb = util.onlyCallOnce(cb)

  const config = context.config
  const proto = config.web.proto

  let addrString
  try {
    addrString = util.normalizeAddress(config.web.server)
  } catch (err) {
    const msg = `error parsing server address ${config.web.server}: ${err}`
    return util.cbNextTick(cb, new Error(msg))
  }

  const addrParts = util.parseAddress(addrString)
  const addrHost = addrParts[0]
  const addrPort = addrParts[1]

  if (context.schemas == null) context.schemas = {}
  if (context.schemas.http == null) context.schemas.http = {}

  const httpSchemas = context.schemas.http
  if (httpSchemas['3'] == null) httpSchemas['3'] = util.readSchema('http-v3')

  qsp.initialize(context)

  const app = express()
  app.disable('x-powered-by')
  app.set('env', 'production')

  app.get('/api', logRequest, (req, res) => {
    res.send({
      versions: {
        program: context.version,
        api: API_VERSIONS
      }
    })
  })

  // routes that we let the client handle
  const webRoutes = [
    '/',
    '/processes',
    '/processes/:id',
    '/assets',
    '/profile/:id',
    '/profile/:id/:nodeid',
    '/snapshot/:id',
    '/snapshotdiff/:idLeft/:idRight',
    '/general-settings',
    '/notifications',
    '/settings',
    '/settings/:id',
    '/security',
    '/vulnerability/:id',
    '/welcome',
    '/welcome/:id'
  ]

  const indexHtml = path.join(__dirname, '..', 'web', 'index.html')

  for (let webRoute of webRoutes) {
    const handlers = [ (req, res) => { res.sendFile(indexHtml) } ]
    handlers.unshift(checkWebRouteLicense.bind(null, context))

    app.get(webRoute, handlers)
    // Since urls to these might be generated by concatenating user input, they
    // potentially might have a double //, so don't break on that.
    const pathPattern = '^/+' + webRoute.replace(/^\//, '')
    app.get(pathPattern, (req, res) => {
      res.redirect(302, webRoute)
      res.end()
    })
  }

  // figure out which of production (index.js.gz) or dev (index.js) to send
  let indexJs = path.join(__dirname, '..', 'web', 'index.js')
  let indexJsGz = `${indexJs}.gz`
  const indexJsHeaders = {
    'Content-Type': 'application/javascript'
  }

  if (util.fileExists(indexJsGz)) {
    Logger.debug('using production version of browser index.js')
    app.get('/index.js', (req, res) => {
      fileSender.sendFile(req, res, indexJsGz, true, indexJsHeaders)
    })
  } else {
    Logger.warn('using development version of browser index.js')
    app.get('/index.js', (req, res) =>
    fileSender.sendFile(req, res, indexJs, false, indexJsHeaders)
    )
  }

  app.use('/', express.static(path.join(__dirname, '..', 'web')))
  app.use('/support/widget', express.static(path.join(__dirname, '..', 'web/support-widget.html')))
  app.use('/support/assets/js', express.static(path.join(__dirname, '..', 'web/support-widget.js')))

  // create a body parser for text that will take ANY content-type, or none
  const bodyCollector = bodyParser.text({type: () => true})

  // periodically log the number of running HTTP requests
  setInterval(() => {
    const eioClients = eioClientCount(context)
    Logger.info(`Handling ${runningQueries.size + eioClients} connected queries.`)
  }, LOG_INTERVAL).unref()

  context.eioHandlers = new Map()

  // iterate through the http schema to get method/uri's to register
  for (let version of API_VERSIONS) {
    for (let method of ['GET', 'PUT', 'POST', 'DELETE']) {
      const methodURIs = httpSchemas[version][method]

      for (let uri in methodURIs) {
        const mod = loadUriModule(version, method, uri)
        if (mod == null) continue

        const expressMethod = method.toLowerCase()
        const urlParm = methodURIs[uri].urlParm
        const needsBody = methodURIs[uri].request != null
        const unlicensedOk = methodURIs[uri].unlicensedOk || false
        const continuous = methodURIs[uri].keepalive || false

        // build up middleware stack
        const middleware = []

        // logger
        middleware.push(logRequest)

        // require a valid license to use the api
        if (!unlicensedOk) {
          middleware.push(checkLicense.bind(null, context))
        }

        // don't kill idle streaming connections, log how many are executing
        middleware.push(accountRequest)

        // if request takes a body, add body collector and parser
        if (needsBody) {
          middleware.push(bodyCollector)
          middleware.push(bodyParserChecker)
        }

        // add the actual module handler for the request at the end
        middleware.push((req, res) => {
          handleAPIRequest(context, mod, version, method, uri, req, res)
        })

        // register primary method / uri handler
        const finalUri = `/api/v${version}/${uri}`
        Logger.debug(`registering URI handler for ${method} ${finalUri}`)
        app[expressMethod](finalUri, middleware)

        // if uri takes a url parameter, register that as well
        if (urlParm) {
          app[expressMethod](`/api/v${version}/${uri}/:${urlParm}`, middleware)
        }

        // build up the handler for engine.io
        if (continuous) {
          context.eioHandlers.set(finalUri, {
            mod: mod,
            uri: finalUri,
            method: method,
            version: version,
            uriSchema: methodURIs[uri],
            unlicensedOk: unlicensedOk
          })
        }
      }
    }
  }

  // add method not found handlers
  for (let version of API_VERSIONS) {
    Logger.debug(`registering not found handlers for /api/v${version}`)
    const methodUriSet = new Set()

    for (let method of ['GET', 'PUT', 'POST', 'DELETE']) {
      const methodUris = httpSchemas[version][method]
      for (let methodUri in methodUris) {
        methodUriSet.add(methodUri)
      }
    }

    for (let uri of methodUriSet) {
      const finalUri = `/api/v${version}/${uri}`
      Logger.debug(`registering URI error handler for * ${finalUri}`)
      app.all(finalUri, (req, res) => {
        res.status(405).send({
          error: `the method ${req.method} is not supported for ${req.originalUrl}`
        })
      })
    }
  }

  // catch all error for any other /api requests
  Logger.debug(`registering URI error handler for * /api/*`)
  app.all(`/api/*`, (req, res) => {
    res.status(404).send({
      error: `the uri ${req.originalUrl} is not a valid API resource`
    })
  })

  const options = {}

  if (proto === 'https') {
    let configName = 'web.https.key'
    let configVal = config.web.https.key

    const keyContents = readHttpsCreds(config, configName, configVal)
    if (keyContents == null) {
      const msg = `unable to read https key file '${configVal}'`
      return util.cbNextTick(cb, new Error(msg))
    }

    configName = 'web.https.cert'
    configVal = config.web.https.cert

    const certContents = readHttpsCreds(config, configName, configVal)
    if (certContents == null) {
      const msg = `unable to read https cert file '${configVal}'`
      return util.cbNextTick(cb, new Error(msg))
    }

    options.key = keyContents
    options.cert = certContents

    if (config.web.https.secureOptionsMask) {
      options.secureOptions = config.web.https.secureOptionsMask
    }
    if (config.web.https.ciphers) {
      options.ciphers = config.web.https.ciphers
    }
  }

  let server

  try {
    if (proto === 'https') {
      server = https.createServer(options, app)
    } else {
      server = http.createServer(app)
    }
  } catch (err) {
    const msg = `error launching server at ${addrString}: ${err}`
    return util.cbNextTick(cb, new Error(msg))
  }

  server.httpAllowHalfOpen = true

  // atttach engine.io server
  context.eioServer = engineIO.attach(server, {
    perMessageDeflate: false
  })
  context.eioServer.on('connection', (socket) => handleEioAPIRequest(context, socket))

  server.on('error', onError)

  server.listen(addrPort, addrHost, serverListening)

  function serverListening (err) {
    if (err) return cb(err)
    const url = util.getHttpBaseURL(context)
    Logger.debug(`calculated server url: ${url}`)
    cb(null, new Server(context, server, app, url))
  }

  function onError (err) {
    if (err.code === 'EADDRINUSE') {
      Logger.error('address of HTTP server already bound by another process', addrString)
    } else {
      Logger.error(err, 'error from http server')
    }

    // because this is an only-call-once cb, will not be called after server
    // is listening
    cb(err)
  }

  context.eventManager.on('license-updated', evt => {
    if (evt.args && evt.args.licensed === false) {
      // If license becomes invalid, tear down any still-running queries that require one
      const cancelErr = new Error('query canceled due to invalid license')
      for (let req of runningQueries) {
        if (req.requiresLicense) {
          Logger.debug(`shutting down http request ${req.method} ${req.url} to to invalid license`)
          req.destroy(cancelErr)
        }
      }
    }
  })
}

// Parse a request body and sanitize.
function bodyParserChecker (req, res, next) {
  req.nsConsole = req.nsConsole || {}

  // When req.body isn't a string, will be {}, indicates error.
  // Can happen with no content-type, etc.
  if (typeof req.body !== 'string') {
    req.nsConsole.bodyErr = new Error('error processing request body')
    return next()
  }

  // parse the body back into itself, but capture error as well
  try {
    req.body = JSON.parse(req.body)
  } catch (err) {
    req.nsConsole.bodyErr = err
    return next()
  }

  // ensure it's an object, not an array, string, etc
  if (!lodash.isPlainObject(req.body)) {
    req.nsConsole.bodyErr = new Error('request body must be a JSON object, not array or primitive')
    return next()
  }

  next()
}

// Handle an HTTP API request.
function handleAPIRequest (context, mod, version, method, uri, req, res) {
  const uriSchema = context.schemas.http[version][method][uri]

  req.nsConsole = req.nsConsole || {}
  req.nsConsole.context = context
  req.nsConsole.uriSchema = uriSchema

  // add urlParm parm to query
  const urlParm = uriSchema.urlParm
  if (urlParm != null) {
    const param = req.params[urlParm]
    if (param != null) {
      req.query[urlParm] = param
    }
  }

  const qData = qsp.parse(context, uriSchema, req.query)
  req.nsConsole.qData = qData

  if (lodash.isString(uriSchema.response)) {
    let match = uriSchema.response.match(/^stream:(.*)/)
    if (match) {
      req.nsConsole.eventStream = EventStream.create(context, req, res)
      req.nsConsole.eventStreamType = match[1]
    }
  }

  if (qData.errors.length !== 0) {
    for (let message of qData.errors) {
      Logger.warn(`error parsing query parameters: ${message}`)
    }

    if (req.nsConsole.eventStream != null) {
      const name = 'error parsing query parameters'
      req.nsConsole.eventStream.writeError(null, name, qData.errors.join('; '))
      req.nsConsole.eventStream.close()
      return
    }

    res.status(400)
    res.send({
      error: {
        message: 'errors occurred parsing query parameters'
      },
      errors: qData.errors
    })

    return
  }

  mod.handleRequest(req, res)
}

// Handle an EngineIO API request.
function handleEioAPIRequest (context, socket) {
  totalQueries++

  const req = socket.request
  const eventStream = EventStream.createEio(context, socket)

  req.nsConsole = req.nsConsole || {}
  req.nsConsole.context = context
  req.nsConsole.eventStream = eventStream

  // url / query string comes in as the only message
  socket.once('message', setupStream)

  function setupStream (data) {
    const errorName = 'engine.io initialization'

    // parse message
    data = data.toString()
    try {
      data = JSON.parse(data)
    } catch (err) {
      const message = `error parsing engine.io JSON message: ${data}`
      eventStream.writeError(null, errorName, message)
      eventStream.close()
      return Logger.error(err, `${errorName}: ${message}`)
    }

    data = data || {}
    const url = data.url
    const queryString = data.queryString || ''

    if (url == null) {
      const message = `message did not contain a url property`
      eventStream.writeError(null, errorName, message)
      eventStream.close()
      return Logger.error(`${errorName}: ${message}`)
    }

    const handlerData = context.eioHandlers.get(url)
    if (handlerData == null) {
      const message = `url does not support streaming: ${url}`
      eventStream.writeError(null, errorName, message)
      eventStream.close()
      return Logger.error(`${errorName}: ${message}`)
    }

    const licenseValid = context.provider && context.provider.license && context.provider.license.valid
    if (!handlerData.unlicensedOk && !licenseValid) {
      const message = `server license is not valid`
      eventStream.writeError(null, errorName, message)
      eventStream.close()
      return
    }

    let queryStringObject
    try {
      queryStringObject = QueryString.parse(queryString)
    } catch (err) {
      const message = `invalid queryString property: ${queryString}`
      eventStream.writeError(null, errorName, message)
      eventStream.close()
      return Logger.error(err, `${errorName}: ${message}`)
    }

    const uriSchema = handlerData.uriSchema
    req.nsConsole.uriSchema = uriSchema

    const qData = qsp.parse(context, uriSchema, queryStringObject)
    if (qData.errors.length !== 0) {
      for (let message of qData.errors) {
        Logger.warn(`error parsing query parameters: ${message}`)
      }

      eventStream.writeError(null, `${errorName}: query parameter errors:`, qData.errors.join('; '))
      eventStream.close()
      return
    }

    req.nsConsole.qData = qData

    const startTime = Date.now()
    const urlLogged = queryString === '' ? url : `${url}?${queryString}`
    Logger.info(`opening engine.io socket on ${urlLogged}`)

    let loggedEnd = false
    socket.on('close', () => logEnd())
    socket.on('error', (err) => logEnd(err))

    handlerData.mod.handleRequest(req, {engineIO: 'requests have no response'})

    // log a message when stream ends
    function logEnd (err) {
      if (loggedEnd) return

      const elapsedMs = Date.now() - startTime
      if (err) {
        Logger.info(err, `closing engine.io socket on ${urlLogged} because of error - ${elapsedMs}ms`)
      } else {
        Logger.info(`closing engine.io socket on ${urlLogged} - ${elapsedMs}ms`)
      }
    }
  }
}

// Load a module that handles a particular HTTP method and uri
function loadUriModule (version, method, uri) {
  const moduleName = (version === '1')
    ? `./http-api/${method}-${uri}`
    : `./http-api/v${version}/${method}-${uri}`
  try {
    return require(moduleName)
  } catch (err) {
    Logger.error(err, `error loading ${moduleName}`)
    return null
  }
}

// Read an HTTPS creds file.
function readHttpsCreds (config, key, relName) {
  let baseName = '.'
  if (config.fileName) baseName = path.dirname(config.fileName)

  Logger.debug(`readHttpsCreds(config, '${key}', '${relName}'), baseName: '${baseName}'`)
  const fileName = path.resolve(baseName, relName)
  Logger.debug(`readHttpsCreds() reading '${fileName}'`)

  const contents = util.loadFile(fileName)
  if (contents == null) {
    Logger.error(`error reading file '${fileName}', specified in config key ${key}`)
    return null
  }

  return contents
}

// Log interesting requests.
function logRequest (req, res, next) {
  const url = req.url
  const method = req.method
  const timeStart = Date.now()

  res.addListener('close', logWhenDone)
  res.addListener('finish', logWhenDone)

  Logger.debug(`==> ${method} ${url}`)

  next()

  function logWhenDone () {
    const timeElapsed = Date.now() - timeStart

    res.removeListener('close', logWhenDone)
    res.removeListener('finish', logWhenDone)

    const statusCode = res.statusCode || '???'
    Logger.info(`${statusCode} ${method} ${url} ${timeElapsed}ms`)
  }
}

function accountRequest (req, res, next) {
  // default socket timeout is 2 minutes, disable it
  req.setTimeout(0)

  // add request to accounting set for logging
  runningQueries.add(req)
  totalQueries++
  req.on('close', () => runningQueries.delete(req))
  req.on('end', () => runningQueries.delete(req))
  req.on('error', () => runningQueries.delete(req))
  next()
}

const ErrorStateToPage = {
  'EXPIRED': '/welcome/license-key-expired',
  'TOKEN_EXPIRED': '/welcome/license-token-expired',
  'INVALID_LICENSE': '/welcome/',
  'INVALID_TOKEN': '/welcome/',
  'NETWORK_ERROR': '/welcome/no-connection',
  'NO_LICENSE': '/welcome/',
  'TOKEN_REQUIRED': '/welcome/no-token',
  'VALIDATING': '/welcome/'
}
function checkWebRouteLicense (context, req, res, next) {
  if (context.provider && context.provider.license && context.provider.license.valid) {
    return next()
  }

  let page = ErrorStateToPage[context.provider.errorState()] || '/welcome/'

  if (req.url === page) return next()

  res.redirect(302, page)
  res.end()
}

function checkLicense (context, req, res, next) {
  if (context.provider && context.provider.license && context.provider.license.valid) {
    // Hang a flag on req to say that this is license-dependent.  If a long
    // running query is still going when the license goes bad, we'll terminate
    // it.
    req.requiresLicense = true

    return next()
  }
  res.status(403)
  res.send({ error: 'Valid license not provided' })
  res.end()
}

function eioClientCount (context) {
  if (context == null || context.eioServer == null) return 0

  return context.eioServer.clientsCount
}

function eioClientShutdown (context) {
  if (context == null || context.eioServer == null) return

  context.eioServer.close()
}

// Contains all the HTTP server goodness.
class Server {
  constructor (context, httpServer, app, url) {
    this.context = context
    this.httpServer = httpServer
    this.app = app
    this.url = url
  }

  queryCount () {
    return totalQueries
  }

  shutdown () {
    // close running queries
    const shutdownErr = new Error('server shutdown')
    for (let req of runningQueries) {
      Logger.debug(`shutting down http request ${req.method} ${req.url}`)
      req.destroy(shutdownErr)
    }

    eioClientShutdown(this.context)

    this.httpServer.close(err => this._onClose(err))
  }

  _onClose (err) {
    this.context = null
    this.httpServer = null
    this.app = null

    if (err) {
      Logger.error(err, 'error closing server')
      return
    }

    Logger.info('server closed')
  }
}
