'use strict'

const lodash = require('lodash')

const api = require('./api')
const util = require('../util')

exports.create = create

// Buffer writes for this many millis.
const BUFFER_FOR_MS = 250
const CHUNK_SIZE = 100

const SkipEventLoggingTypes = new Set([
  'field-range-changed',
  'influxdb-error'
])

// A typical metrics record is about a kilobyte, so let's not keep more than
// 500 MB worth of them
const MAX_QUEUE_WHILE_STARTING = 500000

const Logger = util.getLogger(__dirname, __filename)

// Create and return a new database writer.
function create (context) {
  return new Writer(context)
}

class Writer {
  constructor (context) {
    this._writerMetrics = new QueuedWriter(context, 'agentData')
    this._writerEvents = new QueuedWriter(context, 'rp_events')
  }

  // Write an metrics record to the database.
  writeMetrics (agent, message) {
    // Ensure sane shape of body before modifying it
    const data = lodash.defaults({}, message.body)

    message.body = data
    const line = this._createLine(agent, 'metrics', message)
    this._writerMetrics.writeLine(line)
  }

  // Write a system event record to the database.
  writeEvent (name, agentId, args, time) {
    if (SkipEventLoggingTypes.has(name)) return

    const timestamp = util.uniqDateNS(time)

    const agentIdPart = (agentId == null)
      ? ''
      : `agent=${this.escapeFieldValue(agentId)},`

    if (args === undefined) args = null
    const bodyPart = `body=${this.escapeFieldValue(args)}`
    const typePart = `type=${this.escapeKey(name)}`

    const line = `events,${typePart} ${agentIdPart}${bodyPart} ${timestamp}`
    this._writerEvents.writeLine(line)
  }

  // influx keys and tags must escape "=", " ", and ","
  escapeKey (key) {
    return key.replace(/([ =,])/g, '\\$1')
  }

  // format influx field values
  escapeFieldValue (value) {
    if (value == null) return '"null"'  // return null as the string 'null'
    if (lodash.isNumber(value)) return value
    if (lodash.isBoolean(value)) return `${value}`

    // if it smells like a string, ensure it's a string
    if (lodash.isString(value)) {
      value = `${value}`
    // last chance - it's an object, so JSON stringify
    } else {
      value = JSON.stringify(value)
    }

    // return quoted string, escaping quotes
    return `"${value.replace(/"/g, '\\"')}"`
  }

  _createLine (agent, measurement, message) {
    const tags = Object.keys(agent.tags).length
      ? ',' + lodash.map(agent.tags, (val, key) => `${this.escapeKey(key)}=${this.escapeKey(val)}`).join(',')
      : ''

    const fields = lodash.map(message.body, (val, key) => `${this.escapeKey(key)}=${this.escapeFieldValue(val)}`).join(',')

    const timestamp = message.timeNS || util.uniqDateNS()
    return `${measurement}${tags} ${fields} ${timestamp}`
  }
}

class QueuedWriter {
  constructor (context, rp) {
    this._context = context
    this._rp = rp
    this._timeout = null
    this._queue = []
    this._writing = false
    this._warnedQueueTruncation = false
    this._errorState = false
  }

  writeLine (line) {
    Logger.debug(`queuing line to InfluxDB: ${line}`)
    this._queue.push(line)

    if (this._queue.length >= CHUNK_SIZE) {
      this._writeQueued()
      return
    }

    if (this._timeout) return

    this._timeout = setTimeout(() => this._writeQueued(), BUFFER_FOR_MS)
    this._timeout.unref()
  }

  _writeQueued () {
    // Do this synchronously because data can arrive while influx is writing
    // and we want to allow a new timer to start in the meantime.
    if (this._timeout) clearTimeout(this._timeout)
    this._timeout = null

    if (this._queue.length === 0) return

    // Don't try to write while Influx is still launching.  This assumes that
    // influxdb will be fully running and the db initialized by the time
    // influxdbProcess is saved to Context.  If we add support for external
    // influxdb, this will need to account for that.
    if (!this._context.influxdbProcess && !this._context.usingExternalInflux) {
      if (this._queue.length > MAX_QUEUE_WHILE_STARTING) {
        if (!this._warnedQueueTruncation) {
          Logger.warn(`Truncating metrics because Influx isn't started yet and write queue is full`)
          this._warnedQueueTruncation = true
        }
        this._queue = this._queue.slice(this._queue.length - MAX_QUEUE_WHILE_STARTING)
      }
      return
    }

    // Only write one chunk at a time.  If an extended buffering period has
    // been in place, there could be many thousands of lines to process, so
    // don't stampede.
    if (this._writing) return
    this._writing = true

    Logger.debug(`writing ${this._queue.length} queued line to InfluxDB`)

    // Write the first 100 lines in the queue
    const joinedLines = this._queue.slice(0, CHUNK_SIZE).join('\n')
    this._queue = this._queue.slice(CHUNK_SIZE)

    api.write(this._context.config.influxdb.url, 'nsolid', joinedLines, this._rp, err => {
      this._writing = false
      if (err) {
        // TODO: consider requeuing data if using external influx
        Logger.error(err, 'error while writing metrics to Influx, data discarded')
        if (this._context.eventManager) {
          this._context.eventManager.emitSystemEvent('influxdb-error', null, { message: 'write error' })
        }
        this._errorState = true
        return
      }

      if (this._errorState) {
        if (this._context.eventManager) {
          this._context.eventManager.emitSystemEvent('influxdb-recovered', null, { message: 'write recovery' })
        }
        this._errorState = false
      }
      if (this._queue.length >= CHUNK_SIZE) {
        this._writeQueued()
      }
    })
  }
}
