'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)

const nsolid = require('nsolid')

const validStates = new Set(['ACTIVE', 'INACTIVE'])
const GLOBAL_OFFLINE_MODE = isGlobalOffline()

function isGlobalOffline () {
  if (process.env.NSOLID_CONSOLE_OFFLINE_MODE == null) return false
  switch (process.env.NSOLID_CONSOLE_OFFLINE_MODE.toLowerCase()) {
    case 'false':
    case '0':
      return false
    default:
      return true
  }
}

const noLicense = {
  provider: 'default',
  offlineMode: GLOBAL_OFFLINE_MODE,
  key: '',
  providerData: ''
}
let context

class LicenseKey {
  constructor (provider, existing, _context) {
    context = _context
    if (existing == null) {
      existing = Object.assign({}, noLicense)
      existing.provider = provider
    }
    this.offlineMode = !!existing.offlineMode || GLOBAL_OFFLINE_MODE
    this.key = existing.key
    this.provider = existing.provider || provider
    this.providerData = existing.providerData
    this.firstName = existing.firstName
    this.lastName = existing.lastName
    this.email = existing.email
    this.marketingOptIn = existing.marketingOptIn
    this.company = existing.company
    this.expires = existing.expires
    this.trial = existing.trial
    this.keyState = existing.keyState
    this.token = existing.token || process.env.NSOLID_LICENSE_TOKEN
    this.populateFromToken()
    this.recalculateValidity()
  }

  recalculateValidity () {
    const tokenValid = this.validateToken()
    this.isExpired = !!(this.expires && this.expires < Date.now())
    if (tokenValid && !this.keyState) {
      this.keyState = 'ACTIVE'
    }
    this.valid = !!(tokenValid &&
      (this.offlineMode || (validStates.has(this.keyState) && !this.isExpired)))

    // Should this verify that the license key in the token
    // matches the one in this.key?
  }

  validateToken () {
    if (!this._parsedToken && !this.parseToken()) return

    if (this.key && this.key !== this._parsedToken.licenseKey) {
      Logger.error('Token key does not match license key')
      return false
    }

    return nsolid.validateToken(this.token).licensed
  }

  populateFromToken () {
    if (!this._parsedToken && !this.parseToken()) return
    if (!this.expires) this.expires = this._parsedToken.expires
    if (!this.key) this.key = this._parsedToken.licenseKey
  }

  parseToken () {
    if (!this.token) return
    const parsed = nsolid.parseToken(this.token)
    this._parsedToken = parsed
    return parsed
  }

  save (cb) {
    Logger.debug(`save(${JSON.stringify(this)})`)

    this.recalculateValidity()

    const copy = Object.assign({}, noLicense, this)
    context.settingsManager.setSetting('license', copy, err => {
      if (err) {
        Logger.error(err, 'Could not save license to db.')
        return cb(err)
      }

      Logger.debug('Saved a new license.', this)
      cb(null, this)
    })
  }
}

// creates a LicenseKey from a packet returned from the license service
LicenseKey.fromBody = function fromBody (provider, body, context) {
  if (!context) throw new Error('no context')
  Logger.debug('creating license key from server body', body)
  if (body == null) return new LicenseKey(null, null, context)

  const keyForm = {
    offlineMode: false, // if we're here we must not be in offlineMode...
    provider: body.provider || provider, // allow saved license to override display provider
    providerData: body.providerData,
    firstName: body.name,
    lastName: body.lastname,
    key: body.licensekey,
    email: body.email,
    marketingOptIn: body.marketingOptIn,
    company: body.company,
    expires: body.expires,
    trial: body.trial,
    keyState: body.state,
    token: body.token
  }
  return new LicenseKey(provider, keyForm, context)
}

LicenseKey.GLOBAL_OFFLINE_MODE = GLOBAL_OFFLINE_MODE

module.exports = LicenseKey
