#!/usr/bin/env node

'use strict'

exports.main = main
exports.shutdown = shutdown
exports.getConfig = getConfig

const HttpProxyAgent = require('http-proxy-agent')
const HttpsProxyAgent = require('https-proxy-agent')

const pkg = require('../package.json')
const util = require('./util')
const Config = require('./config')
const SettingsManager = require('./settings-manager')

// change 0 to something like 30000 to test long influx startup

const Logger = util.getLogger(__filename)

const Context = {}

function main (cb) {
  // try loading zmq early, log error if can't be loaded
  try {
    util.requireZmq()
  } catch (err) {
    Logger.error(err, 'error loading zmq')
    Logger.error('this program should be run with the N|Solid runtime, or you can manually npm install zmq for diagnostic purposes')
    process.exit(1)
  }

  process.on('SIGINT', shutdownHandler('SIGINT', 0))
  process.on('SIGTERM', shutdownHandler('SIGTERM', 0))
  process.on('uncaughtException', shutdownHandler('uncaughtException', 1))

  Context.version = pkg.version
  Context.processStart = Date.now()
  Context.isStaging = (process.env.NSOLID_STAGING != null)
  Context.mainCB = cb
  Context._running = false
  Context._onceRunningCallbacks = []

  Context.onceRunning = cb => {
    if (Context._running) {
      setImmediate(cb)
    } else {
      Context._onceRunningCallbacks.push(cb)
    }
  }

  if (!Context.config) {
    Context.config = getConfig()
  }

  // repairing the DB runs async, so need to stop normal server startup here
  if (Context.config.do_repairDb) return

  // wait till help, etc have had a chance to run before adding exit handler
  process.on('exit', onExit)

  if (Context.isStaging) Logger.info('using staging')

  Logger.debug('config:', Context.config)

  Logger.info(`N|Solid Console Service starting up with PID ${process.pid}`)

  setupHttpAgent(Context)

  if (Context.config.relayMode) return startRelay()

  return startConsoleServer()
}

function startRelay () {
  Logger.info('launching in Relay mode')

  // Lazy-load Relay
  const zmqRelay = require('./boot/relay')

  zmqRelay.launch(Context, Config)
    .then(() => allLaunched())
    .catch(launchError)

  return Context
}

function startConsoleServer () {
  // Lazy load so that we don't load lots of unnecessary things for the Relay
  const boot = require('./boot')
  const managers = require('./boot/managers')

  managers.register(Context)

  boot.settings.launch(Context)
    .then(() => {
      return new Promise((resolve, reject) => {
        managers.registerSettingsDependant(Context, err => {
          if (err) return reject(err)
          boot.influxdb.createClients(Context)
          resolve()
        })
      })
    })
    .then(() => {
      return Promise.all([
        boot.zmqServer.launch(Context),
        boot.influxdb.launch(Context),
        boot.httpServer.launch(Context)
      ])
    })
    .then(() => allLaunched())
    .catch(launchError)

  return Context
}

// Run some code once all subsystems have started
function allLaunched () {
  Logger.info('all subsystems launched!')

  // Relay doesn't have an eventManager
  if (Context.eventManager) {
    Context.eventManager.emitSystemEvent('console-server-started')
  }

  Context._running = true
  if (Context._onceRunningCallbacks && Context._onceRunningCallbacks.length) {
    Context._onceRunningCallbacks.forEach(cb => {
      cb()
    })
  }

  if (Context.mainCB) return Context.mainCB(null, Context)
}

function launchError (err) {
  Logger.error(err)
  if (Context.mainCB) return setImmediate(Context.mainCB, err)
  setImmediate(shutdown, 1)
}

function shutdownHandler (name, code) {
  return (err) => {
    // ignore errors on stdout, while going down ...
    Logger.ignoreStreamError(true)
    Logger.error(`stopping process; reason: ${name}`)
    if (err && err instanceof Error) Logger.error(err)

    shutdown(code, err)
  }
}

const INFLUX_SHUTDOWN_MAX_SECONDS = 30
let ShuttingDown = false

function shutdown (code, err) {
  // ignore errors on stdout, while going down ...
  Logger.ignoreStreamError(true)

  if (ShuttingDown) {
    Logger.error(`shutdown(${code}) called while already shutting down`)
    return
  }

  ShuttingDown = true

  // Relay doesn't have an eventManager
  if (Context.eventManager) {
    Context.eventManager.emitSystemEvent('console-server-stopped', null, {
      exitCode: code,
      message: err ? err.message : undefined,
      stack: err ? err.stack : undefined
    })
  }

  if (Context.zmqRelay) {
    Context.zmqRelay.shutdown()
    Context.zmqRelay = null
  }
  if (Context.zmqServer) {
    Context.zmqServer.shutdown()
    Context.zmqServer = null
  }
  if (Context.httpServer) {
    Context.httpServer.shutdown()
    Context.httpServer = null
  }
  if (Context.diskAssetManager) {
    Context.diskAssetManager.shutdown()
    Context.diskAssetManager = null
  }
  if (Context.vulnManager) {
    Context.vulnManager.shutdown()
    Context.vulnManager = null
  }

  if (Context.usingExternalInflux) {
    // We can't stop external influx, so we're ready to exit now.
    process.exit(0)
  }

  // Last stop before having to shutdown influx.  If it was started but didn't
  // complete startup, it may have the influxdbLaunchingProcess property.
  const influx = Context.influxdbProcess || Context.influxdbLaunchingProcess
  if (!influx) {
    Logger.error('InfluxDB did not successfully start; exiting')
    process.exit(code)
  }

  // influxd can take a few seconds to come down, so gate actual exit on it
  Logger.info('waiting for InfluxDB process to shutdown')
  influx.kill(() => {
    Logger.info('InfluxDB process shutdown')
    process.exit(code)
  })

  setTimeout(() => {
    Logger.error('forcing shutdown, InfluxDB not finished shutting down')
    process.exit(1)
  }, INFLUX_SHUTDOWN_MAX_SECONDS * 1000).unref()
}

function onExit (code) {
  // ignore errors on stdout, while going down ...
  Logger.ignoreStreamError(true)
  Logger.info(`main process exiting with code ${code}`)
}

// Set the global http client agent, with HTTP_PROXY override.
function setupHttpAgent (context) {
  const config = context.config

  const httpProxy = config.web.httpProxy
  const httpsProxy = config.web.httpsProxy

  if (httpProxy != null) {
    Logger.info(`using HTTP proxy: ${httpProxy}`)
    context.httpAgent = new HttpProxyAgent(httpProxy)
  }

  if (httpsProxy != null) {
    Logger.info(`using HTTPS proxy: ${httpsProxy}`)
    context.httpsAgent = new HttpsProxyAgent(httpsProxy)
  }
}

// get the config, handling -h and -v inline
function getConfig () {
  let conf
  try {
    conf = Config.getConfig()
  } catch (e) {
    Logger.error(e, 'Error attempting to load configuration.')
    process.exit(1)
  }

  // if returned configuration is null, a logged error already printed
  if (conf == null) {
    Logger.error('Failed to load configuration, stopping.')
    process.exit(1)
  }

  if (conf.do_help) {
    util.printHelp()
    process.exit(0)
  }

  if (conf.do_version) {
    console.log(pkg.version)
    process.exit(0)
  }

  if (conf.do_genkey) {
    genCurveKeypair()
    process.exit(0)
  }

  if (conf.do_repairDb) {
    repairDb(conf, err => {
      // success/failure message already written
      if (err) process.exit(1)
      process.exit(0)
    })
    return conf
  }

  if (conf.do_configPrint) {
    delete conf.do_configPrint
    console.log(JSON.stringify(conf, null, 4))
    process.exit(0)
  }

  return conf
}

function repairDb (config, cb) {
  SettingsManager.repairDb(config, cb)
}

function genCurveKeypair () {
  const zmq = util.requireZmq()
  const pair = zmq.curveKeypair()

  process.stdout.write('#################################\n')
  process.stdout.write('## N|Solid Console config TOML ##\n')
  process.stdout.write('#################################\n')
  process.stdout.write('[sockets]\n')
  process.stdout.write('\n')
  process.stdout.write(`publicKey = "${pair.public}"\n`)
  process.stdout.write(`privateKey = "${pair.secret}"\n`)
  process.stdout.write('\n')
  process.stdout.write('######################################\n')
  process.stdout.write('## N|Solid Console environment vars ##\n')
  process.stdout.write('######################################\n')
  process.stdout.write(`NSOLID_CONSOLE_SOCKETS_PUBLIC_KEY='${pair.public}'\n`)
  process.stdout.write(`NSOLID_CONSOLE_SOCKETS_PRIVATE_KEY='${pair.secret}'\n`)
  process.stdout.write('\n')
  process.stdout.write('#####################################\n')
  process.stdout.write('## N|Solid Runtime environment var ##\n')
  process.stdout.write('#####################################\n')
  process.stdout.write(`NSOLID_PUBKEY='${pair.public}'\n`)
}

if (require.main === module) {
  process.title = 'nsolid-console'
  main()
}
