'use strict'

// const request = require('client-request')
const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const crypto = require('crypto')
const lodash = require('lodash')
const retry = require('./retry-request')
const querystring = require('querystring')
const LicenseKey = require('./license-key')
const RevocableEventEmitter = require('./RevocableEventEmitter')
const nsolid = require('nsolid')

let provider
// Number of minutes to cache license from license service
const LICENSE_CHECK_MINS = 30

const LICENSE_EXPIRY_CHECK_SEC = 60

const LICENSE_FETCH_TIMEOUT = 10000
const LICENSE_CHECK_TIMEOUT = 3000
const REPORT_TIMEOUT = 60000

module.exports.create = create

// Requires that settings have been loaded from db
function create (context, cb) {
  context.settingsManager.getSettings('license', (err, licenseInfo) => {
    if (err) return setImmediate(cb, err)

    let configProvider = null

    // always set in config, value of --provider cli option, lowercased;
    // if not passed on cli, will be "default"
    let cliProvider = lodash.get(context, 'config.license.provider')

    // pick provider in license.json over cli option
    if (licenseInfo && licenseInfo.provider) {
      configProvider = licenseInfo.provider
    } else {
      configProvider = cliProvider
    }

    configProvider = configProvider.toLowerCase()

    // if provider was specified in config (CLI option), but we ended up
    // using the provider in the license, issue a warning
    if (configProvider !== cliProvider) {
      Logger.warn(`using saved provider "${configProvider}" instead of requested provider "${cliProvider}"`)
    }

    // already have that provider, so return it
    if (provider && provider.type === configProvider) {
      return setImmediate(cb, null, provider)
    }

    // create a new provider
    if (configProvider === 'azure') {
      provider = new AzureProvider({ context: context })
    } else if (configProvider === 'aws') {
      provider = new AWSProvider({ context: context })
    } else if (configProvider === 'gcp') {
      provider = new GCPProvider({ context: context })
    } else if (configProvider === 'homebrew') {
      provider = new HomebrewProvider({ context: context })
    } else if (configProvider === 'default') {
      provider = new DefaultProvider({ context: context })
    } else {
      return setImmediate(cb, new Error('Provider misconfigured'))
    }

    provider.expireCheckInterval = setInterval(() => {
      if (provider.license && provider.license.valid &&
          provider.license.expires && provider.license.expires < Date.now()) {
        // Force a re-check to update valid flag
        provider.license.recalculateValidity()

        // Notify agents that we're now unlicensed.
        // They should already know this, but send it anyway.
        provider.broadcastLicenseState()
      }
    }, LICENSE_EXPIRY_CHECK_SEC * 1000).unref()

    cb(null, provider)
  })
}

// EventEmitter only used for tests.
class BaseProvider extends RevocableEventEmitter {
  constructor (params) {
    super()
    this._context = params.context
    this.canOptOut = LicenseKey.GLOBAL_OFFLINE_MODE
    this.keyViaEmail = false
    this.url = this._context.isStaging
      ? 'https://staging-services.nodesource.io/v2/licenses'
      : 'https://services.nodesource.io/v2/licenses'
    this.reportUrl = this._context.isStaging
      ? 'https://staging-services.nodesource.io/v2/report'
      : 'https://services.nodesource.io/v2/report'

    this.timer = null
    this.ready = null
    this._checkError = null

    this._previouslyBroadcastLicense = null
    this._previouslyBroadcastErrorState = null
  }

  init () {
    this.getLicenseKeyArgs((err, licenseKeyArgs) => {
      if (err) {
        Logger.error('Cannot look up license information.', err)
      } else if (licenseKeyArgs) {
        this.license = new LicenseKey(this.type, licenseKeyArgs, this._context)
        this.broadcastLicenseState()
      } else {
        Logger.info('No license configured on load.')
      }

      // Notify any agents already running what the current license situation is,
      // they may be waiting for authorization.  Wait for the server to complete
      // startup if necessary.
      this._context.onceRunning(() => {
        this._context.zmqServer.sendConfig('broadcast')
      })

      if (this.isOffline()) {
        Logger.debug('License indicates offlineMode, so no license checking')
        return
      }

      this.startTimer()
      // This will test the validity of the license
      setImmediate(this.refreshLicense.bind(this))
    })
  }

  // License or token passed by command line or env var takes precedence over
  // any license information saved in the database.  In order to avoid warnings
  // about not having a valid token, we should make a best effot to use one if
  // we have it.
  getLicenseKeyArgs (cb) {
    this._context.settingsManager.getSettings('license', (err, savedLicense) => {
      if (err) return cb(err)

      // 1. If a token is passed via env, use it
      if (process.env.NSOLID_LICENSE_TOKEN) {
        return cb(null, {
          token: process.env.NSOLID_LICENSE_TOKEN
        })
      }

      // 2. If a license key is passed via env or config, use it
      if (this._context.config.license && this._context.config.license.key) {
        const args = {
          key: this._context.config.license.key
        }

        // 2a.  If a token is persisted and its key matches the passed key, use it
        if (savedLicense && savedLicense.token) {
          const parsed = this.parseToken(savedLicense.token)
          if (parsed.licenseKey === args.key) {
            args.token = savedLicense.token
          }
        }

        return cb(null, args)
      }

      // 3. If nothing is passed but a license key is persisted, use it
      if (savedLicense && savedLicense.key) {
        return cb(null, savedLicense)
      } else {
        return cb(new Error('Unable to find any license.'))
      }
    })
  }

  isOffline () {
    if (LicenseKey.GLOBAL_OFFLINE_MODE) return true
    return this.canOptOut && this.license && this.license.offlineMode
  }

  startTimer () {
    this.timer = setInterval(() => {
      this.refreshLicense()
      this.reportUsage()
    }, LICENSE_CHECK_MINS * 60 * 1000)
    this.timer.unref()
    Logger.debug(`Checking license every ${LICENSE_CHECK_MINS} minutes`)
  }

  saveKey (cb) {
    if (!this.canOptOut) {
      this.license.offlineMode = false
    }
    this.license.save(cb)
  }

  parseToken (token) {
    return nsolid.parseToken(token) || {}
  }

  resetLicense () {
    // Used to force clear a license key
    this.license = new LicenseKey(this.type, null, this._context)

    // Notify agents that we're now unlicensed
    this.broadcastLicenseState()
  }

  refreshLicense (cb) {
    cb = cb || function () {}
    Logger.debug('Beginning refresh license.')

    if (this.isOffline()) {
      if (!this.ready) {
        this.ready = true
        this.emit('ready')
      }
      return cb(null, this.license)
    }

    this.syncLicense((err, license) => {
      if (err) {
        Logger.warn(err, 'Could not refresh license.', license)
        this.broadcastLicenseState()
        return cb(null, this.license)
      }

      this.license = new LicenseKey(this.type, license, this._context)
      this.broadcastLicenseState()

      this.saveKey((err) => {
        if (err) {
          Logger.warn(err, 'Could not save refreshed license.')
        }
        if (!this.ready) {
          this.ready = true
          this.emit('ready')
        }
        return cb(null, this.license)
      })
    })
  }

  reportUsage (cb) {
    cb = cb || function () {}

    if (this.isOffline()) {
      return cb()
    }

    if (!this._context.agentManager || !this._context.httpServer || !this.license) {
      return cb()
    }

    const params = {
      reportTime: Date.now(),
      consoleId: nsolid.id,
      consoleUptime: process.uptime(),
      licenseKey: this.license.key,
      agentHours: this._context.agentManager.report.agentHours(),
      agentActiveHours: this._context.agentManager.report.agentActiveHours(),
      numProcesses: this._context.agentManager.count(),
      medianAgentUptime: this._context.agentManager.report.medianUptime(),
      numHosts: this._context.agentManager.report.numHosts(),
      numCores: this._context.agentManager.report.numCores(),
      numApplications: this._context.agentManager.report.numApplications(),
      versions: {
        console: this._context.version,
        agents: this._context.agentManager.report.versions()
      },
      apiCalls: this._context.httpServer.queryCount()
    }

    const options = {
      uri: this.reportUrl,
      method: 'POST',
      body: params,
      json: true,
      timeout: REPORT_TIMEOUT
    }

    retry(this._context, options, (err, res, body) => {
      if (err) return cb(err)
      if (body && body.errorMessage) {
        // The Lambda returns JSON-within-JSON in the event of body.errorMessage:
        // { errorMessage: '{"status":404,"description":"Not found","detail":"License not found"}' }
        let serverError = errorMessageToObject(body)
        return cb(serverError)
      }
      if (res.statusCode !== 200) {
        return cb(new Error(`http status ${res.statusCode} reporting usage`))
      }
      cb()
    })
  }

  validateKey (key, cb) {
    Logger.debug(`validateKey ${key}`)
    if (this.isOffline()) {
      // It's unlikely that `this.license.offlineMode` is set here, but checking nonetheless
      Logger.warn('refusing to check key due to offlineMode')
      return setImmediate(() => {
        cb(new Error('cannot check key when offlineMode=true'))
      })
    }
    const keysQs = 'tokenKeyIds=' + nsolid.tokenKeys().join(',')
    const options = {
      uri: `${this.url}/${key}?${keysQs}`,
      method: 'GET',
      json: true,
      retryAttempts: 3,
      timeout: LICENSE_CHECK_TIMEOUT
    }

    Logger.debug('Checking license.', options)
    retry(this._context, options, (err, res, body) => {
      if (err) {
        this._checkError = true
        this.broadcastLicenseState()
        return cb(err)
      }

      // If we didn't get a socket error, any 2xx or 4xx is not a check error
      this._checkError = (res.statusCode >= 500)
      this.broadcastLicenseState()

      if (body && body.errorMessage) {
        // The Lambda returns JSON-within-JSON in the event of body.errorMessage:
        // { errorMessage: '{"status":404,"description":"Not found","detail":"License not found"}' }
        let serverError = errorMessageToObject(body)
        return cb(serverError)
      }
      return cb(null, LicenseKey.fromBody(this.type, body, this._context))
    })
  }

  // fetches license from server and saves
  syncLicense (params, cb) {
    Logger.debug(`syncLicense(${JSON.stringify(params)})`)
    if (typeof params === 'function') {
      cb = params
      params = {}
    }

    if (params.key == null && params.email == null && this.license != null) {
      params.key = this.license.key
    }

    // if nothing to check, return current license
    if (!params.email && !params.key) {
      setImmediate(() => { cb(null, this.license) })
      return
    }

    let url = this.url
    const headers = {}

    if (params.key) {
      const keysQs = 'tokenKeyIds=' + nsolid.tokenKeys().join(',')
      url = `${url}/${params.key}?${keysQs}`
    } else {
      headers.Authorization = formatAuthHeaderEmail(params.email)

      const qparms = querystring.stringify({
        provider: this.type,
        providerData: params.providerData || '',
        tokenKeyIds: nsolid.tokenKeys().join(',')
      })

      url = `${url}?${qparms}`
    }

    const options = {
      uri: url,
      method: 'GET',
      json: true,
      headers: headers,
      retryAttempts: 5,
      timeout: LICENSE_FETCH_TIMEOUT
    }

    Logger.debug('Fetching license.', options)
    retry(this._context, options, (err, res, body) => {
      if (err) {
        this._checkError = true
        if (params.notify) {
          this.broadcastLicenseState()
        }
        return cb(err)
      }

      // If we didn't get a socket error, any 2xx or 4xx is not a check error
      this._checkError = (res.statusCode >= 500)

      if (body && body.errorMessage) {
        // The Lambda returns JSON-within-JSON in the event of body.errorMessage:
        // { errorMessage: '{"status":404,"description":"Not found","detail":"License not found"}' }
        let serverError = errorMessageToObject(body)
        if (params.notify) {
          this.broadcastLicenseState()
        }
        return cb(serverError)
      }

      if (res.statusCode === 404) {
        // 404 means successful lookup but no license found by this key
        this.license = new LicenseKey(this.type, {provider: this.type, key: this.license.key}, this._context)
        if (params.notify) {
          this.broadcastLicenseState()
        }

        this.saveKey(cb)
        return
      }
      if (res.statusCode !== 200) {
        return cb(new Error(`http status ${res.statusCode} fetching license`))
      }

      this.license = LicenseKey.fromBody(this.type, body, this._context)

      // Only send this event if asked, because the caller may decide to do
      // this (such as refreshLicense)
      if (params.notify) {
        this.broadcastLicenseState()
      }
      this.saveKey(cb)
    })
  }

  setToken (token) {
    this.license = new LicenseKey(this.type, { token }, this._context)
    this.broadcastLicenseState()
  }

  setOfflineMode (offlineMode, cb) {
    if (!this.canOptOut) {
      // not allowed to offlineMode
      if (offlineMode) {
        Logger.warn(`Ignoring offlineMode request for provider ${this.type}`)
        offlineMode = false
      }
      if (this.license.offlineMode) {
        Logger.warn(`Overriding offlineMode back to false for provider ${this.type}`)
        this.license.offlineMode = false
        return this.saveKey(cb)
      }
    }
    if (this.license.offlineMode === offlineMode) {
      // nothing to do...
      return setImmediate(() => {
        cb(null, this.license)
      })
    }
    this.license.offlineMode = offlineMode
    this.saveKey(cb)
  }

  createLicense (params, cb) {
    Logger.debug(`createLicense(${JSON.stringify(params)})`)

    if (!params.email) {
      setImmediate(() => {
        cb(new Error('Parameter `email` is required to create a license'))
      })
      return
    }

    if (params.providerData == null) {
      params.providerData = ''
    }

    params.tokenKeyIds = nsolid.tokenKeys()

    const options = {
      uri: this.url,
      method: 'POST',
      json: true,
      body: params,
      headers: {
        Authorization: formatAuthHeaderCreator(this._context.version)
      },
      retryAttempts: 5,
      timeout: LICENSE_FETCH_TIMEOUT
    }

    retry(this._context, options, (err, res, body) => {
      if (err) {
        this._checkError = true
        this.broadcastLicenseState()
        return cb(err)
      }

      // If we didn't get a socket error, any 2xx or 4xx is not a check error
      this._checkError = (res.statusCode >= 500)

      if (res.statusCode < 200 || res.statusCode > 202) {
        this.broadcastLicenseState()
        return cb(new Error(res.statusCode))
      }
      if (body == null) {
        this.broadcastLicenseState()
        return cb(new Error('no body returned from createLicense'))
      }
      if (body.errorMessage) {
        // The Lambda returns JSON-within-JSON in the event of body.errorMessage:
        // { errorMessage: '{"status":404,"description":"Not found","detail":"License not found"}' }
        let serverError = errorMessageToObject(body)
        this.broadcastLicenseState()
        return cb(serverError)
      }

      if (this.keyViaEmail) {
        this.resetLicense()
      } else {
        this.license = LicenseKey.fromBody(this.type, body, this._context)
        this.broadcastLicenseState()
      }
      this.saveKey(cb)
    })
  }

  broadcastLicenseState () {
    // Memoize the last state we sent so that we don't inundate the event
    // stream and agents when nothing has actually changed.
    const errorState = this.errorState()
    if (!lodash.isEqual(this._previouslyBroadcastLicense, this.license) || errorState !== this._previouslyBroadcastErrorState) {
      if (this._context.zmqServer) {
        this._context.zmqServer.sendConfig('broadcast')
      }

      // License the console's own runtime so GET-server-info works
      if (this.license && this.license.token) {
        nsolid.start({licenseToken: this.license.token})
      }

      this._context.eventManager.emitSystemEvent('license-updated', null, {
        licensed: this.license.valid,
        provider: this.type,
        errorState
      })

      this._previouslyBroadcastLicense = this.license
      this._previouslyBroadcastErrorState = errorState
    }
  }

  errorState () {
    if (this.license && this.license.valid) return

    // If we've been trying to contact the license server and erroring
    if (this._checkError) return 'NETWORK_ERROR'

    if (this.license) {
      if (this.license.isExpired) {
        // If we are not offline, the expiration is validated by our license
        // server, so only indicate it's a token-related expiry if they're in
        // offline mode.
        if (this.license.token && this.isOffline()) return 'TOKEN_EXPIRED'

        return 'EXPIRED'
      }

      // Have unexpired token, must be invalid
      if (this.license.token) return 'INVALID_TOKEN'

      // No key and no token = no license
      if (!this.license.key) return 'NO_LICENSE'

      // I have a key but no token and in offline mode, token required
      if (this.isOffline()) return 'TOKEN_REQUIRED'

      // No check error means we communicated with the license server, so the
      // license must be invalid
      if (this._checkError === false) return 'INVALID_LICENSE'

      // We have a key and are not in offline mode but it's not valid yet, must
      // be checking
      return 'VALIDATING'
    } else {
      if (this.isOffline()) return 'TOKEN_REQUIRED'

      return 'NO_LICENSE'
    }
  }
}

class DefaultProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'default'
    this.canOptOut = true
    this.keyViaEmail = true
    this.init()
  }
}

class AWSProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'aws'
    this.canOptOut = true
    this.keyViaEmail = true
    this.init()
  }
}

class AzureProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'azure'
    this.canOptOut = true
    this.keyViaEmail = true
    this.init()
  }
}

class GCPProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'gcp'
    this.canOptOut = true
    this.keyViaEmail = true
    this.init()
  }
}

class HomebrewProvider extends BaseProvider {
  constructor (params) {
    super(params)
    this.type = 'homebrew'
    this.canOptOut = true
    this.keyViaEmail = true
    this.init()
  }
}

function formatAuthHeaderEmail (email) {
  const hashedEmail = crypto.createHash('sha256')
    .update(email)
    .digest('hex')

  const authString = Buffer.from(`${hashedEmail}:NodeSource`).toString('base64')
  return `Basic ${authString}`
}

function formatAuthHeaderCreator (version) {
  const authString = Buffer.from(`nsolid-console ${version}:NodeSource`).toString('base64')
  return `Basic ${authString}`
}

function errorMessageToObject (body) {
  let err
  try {
    const errorBody = JSON.parse(body.errorMessage)
    err = new Error(errorBody.detail)
    err.status = errorBody.status
  } catch (parseError) {
    err = new Error(`Server error: ${body.errorMessage}`)
  }
  return err
}
