'use strict'

const util = require('./util')
const Logger = util.getLogger(__dirname, __filename)
const lodash = require('lodash')

exports.create = create
exports.repairDb = repairDb
exports.buildKey = buildKey

const ALL_VIEW = {
  // API needs a UUID. It doesn't matter what it is,
  // but it needs to be linkable.
  id: '4dbe5222-8e54-448a-9d4c-9bba7706b71b',
  expression: [],
  name: 'All Processes',
  persistent: true,
  axisX: 'cpuPercent',
  axisY: 'heapUsed'
}
const VIEWS_KEY = 'savedViews'

const noEventTypes = ['assets', 'agents']

function create (context) {
  return new SettingsManager(context)
}

// Manages the application settings.
class SettingsManager {
  constructor (context) {
    this._context = context

    let conn
    if (context.config.data.postgresDsn) {
      Logger.info('using PostGres')
      conn = require('./connector-postgres')
    } else {
      Logger.info('using LevelDB')
      conn = require('./connector-leveldb')
    }

    this._db = conn.create(context)

    this._db.once('connect', _ => {
      process.nextTick(_ => {
        this.readSettings((err, settings) => {
          if (err) {
            return Logger.warn('Failed to read settings on db connect')
          }

          if (settings.generalSettings &&
              settings.generalSettings.generalSettingsMain &&
              settings.generalSettings.generalSettingsMain.didFirstSetup) return

          this.setDefaultViews().then(
            result => { // Resolve
              Logger.info('Set preset views')
              const settingsUpdate = Object.assign({}, settings.generalSettings, {
                id: 'generalSettingsMain',
                didFirstSetup: true
              })
              this.setSetting('generalSettings', settingsUpdate, err => {
                if (err) {
                  Logger.warn('Failed to complete first-time setup', err)
                  return
                }
                Logger.info('Finished first-time setup')
              })
            },
            reason => { // Reject
              Logger.warn('Failed to set preset views', reason)
            }
          )
        })
      })
    })
  }

  // Returns the agents stored in leveldb, in the callback.
  // The callback is invoked for every agent returned from leveldb.
  // On error, the callback is invoked with the error from leveldb, and
  // no more agents will be returned.
  // When there are no more agents to be returned, a null will be passed
  // to the callback instead of an agent.
  streamAgents (cb) {
    this._db.streamAgents(cb)
  }

  // Return a subset of the current settings.
  getSettings (type, cb) {
    this._db.getSettings(type, (err, res) => {
      if (err) return cb(err)
      if (type === 'savedViews') {
        const viewId = Object.keys(res).find(key => res[key].name === ALL_VIEW.name)
        if (!viewId) {
          res[ALL_VIEW.id] = ALL_VIEW
        }
      }
      cb(null, res)
    })
  }

  // Get a single entry by ID
  getEntry (type, id, cb) {
    if (id === ALL_VIEW.id && type === 'savedViews') return ALL_VIEW
    return this._db.getEntry(type, id, cb)
  }

  readSettings (cb) {
    if (lodash.get(this, '_context.config.data.postgresDsn')) {
      return setImmediate(cb)
    }
    this._db.readSettings(cb)
  }

  deleteSetting (type, id, cb) {
    this._db.deleteSetting(type, id, err => {
      if (err) return cb(err)
      this.emitSettingsEvent(type, {type, id})
      cb()
    })
  }

  emitSettingsEvent (type, args) {
    if (noEventTypes.indexOf(type) !== -1) return
    this._context.eventManager.emitSystemEvent(`${type}-settings-changed`, null, args)
  }

  setSetting (type, item, cb) {
    this._db.setSetting(type, item, (err, item) => {
      if (err) return cb(err)
      this.emitSettingsEvent(type, {type, item})
      cb(null, item)
    })
  }

  setDefaultViews () {
    // Set up preset views
    const { views } = util.readSchema('preset-views')
    if (!views || !Array.isArray(views)) {
      Logger.warn(new Error('Failed to read preset views data'))
      return
    }

    const settingsManager = this

    function makeViewPromise (view) {
      return new Promise((resolve, reject) => {
        settingsManager.setSetting(VIEWS_KEY, view, (err, item) => {
          if (err) {
            Logger.warn(err)
            return reject(err)
          }
          resolve()
        })
      })
    }

    const mappedViews = views.map(view => makeViewPromise(view))
    return Promise.all(mappedViews)
  }
}

/*
 * Different types of data use a different format for their key.
 *
 * Currently we only have one setting with a key that is a static string.
 * There can only be one license, so there is no use having `license:${key}`
 */
function buildKey (config, type, item) {
  let conn
  if (config.data.postgresDsn) {
    conn = require('./connector-postgres')
  } else {
    conn = require('./connector-leveldb')
  }

  return conn.buildKey(type, item)
}

// attempt to repair the leveldb bits, call cb with error on failure
function repairDb (config, cb) {
  let conn
  if (config.data.postgresDsn) {
    // Don't repair postgres
    Logger.warn('Refusing to repair postgres database in SaaS mode.')
    return setImmediate(cb)
  } else {
    conn = require('./connector-leveldb')
  }

  conn.repairDb(config, cb)
}
