'use strict'

const RevocableEventEmitter = require('./RevocableEventEmitter')

const Triejs = require('triejs-but-without-foounit-as-a-prod-dep')

class TimedCache extends RevocableEventEmitter {
  constructor (ttlMs, gcIntervalMs) {
    super()
    if (ttlMs >>> 0 !== ttlMs) {
      throw new Error('TTL should be an integer number of milliseconds')
    }
    this._ttlMs = ttlMs
    this._gcIntervalMs = gcIntervalMs || 1000
    this._cache = new Map()
    this._expiryLookup = new Map()
    // Triejs cache is not useful and prevents deletion; disable it
    this._trie = new Triejs({ enableCache: false })
    this.timerLoop = setInterval(this.gc.bind(this), this._gcIntervalMs)
    this.timerLoop.unref()
  }

  purge () {
    this._cache = new Map()
    this._trie = new Triejs({ enableCache: false })
    this._expiryLookup = new Map()
  }

  upsert (key, entry) {
    // add key, (re)set timer
    let added = !this._cache.has(key)
    this._cache.set(key, entry)
    this._expiryLookup.set(key, Date.now() + this._ttlMs)
    if (added) {
      // triejs allows dupes, so only add this the first time
      this._trie.add(key)
      this.emit('added', key, entry)
    }
  }

  get (possiblyShortKey) {
    const fullKey = this._getFullKey(possiblyShortKey)
    if (!fullKey) return
    this.checkExpired(fullKey)
    return this._cache.get(fullKey)
  }

  has (possiblyShortKey) {
    const fullKey = this._getFullKey(possiblyShortKey)
    if (!fullKey) return
    return this._cache.has(fullKey)
  }

  delete (possiblyShortKey) {
    const fullKey = this._getFullKey(possiblyShortKey)
    var value = this._cache.get(fullKey)
    if (!value) return
    this._delete(fullKey)
    this.emit('deleted', fullKey, value)
  }

  _delete (key) {
    this._cache.delete(key)
    this._expiryLookup.delete(key)
    this._trie.remove(key)
  }

  _getFullKey (possiblyShortKey) {
    const entries = this._trie.find(possiblyShortKey)
    if (entries && entries.length === 1) {
      return entries[0]
    }
  }

  entries () {
    this.gc()
    var clone = {}
    for (var entry of this._cache) {
      clone[entry[0]] = entry[1]
    }
    return clone
  }

  count () {
    return this._cache.size
  }

  checkExpired (key) {
    if (Date.now() > this._expiryLookup.get(key)) {
      this.emit('expired', key, this._cache.get(key))
      this._delete(key)
    }
  }

  gc () {
    this._expiryLookup.forEach((val, key) => {
      this.checkExpired(key)
    })
  }

  stopTimerLoop () {
    clearInterval(this.timerLoop)
  }
}

module.exports = TimedCache
