'use strict'

const zmqzap = require('zmq-zap')
const zap = new zmqzap.ZAP()
const util = require('../util')
const ZMQBindSocket = require('../zmqserver/socket')
const ZMQRemoteSocket = require('./remote-socket')
const BufferQueue = require('./buffer-queue')
const SharedMemoryManager = require('./shared-memory-manager')
const websocket = require('./websocket')

const Logger = util.getLogger(__dirname, __filename)

zap.use(new zmqzap.CurveMechanism(function (data, callback) {
  // Authorize all ZeroMQ socket connections.
  callback(null, true)
}))

class ZeroMQRelay {
  constructor (config, context) {
    this.config = config
    this.context = context

    // We'll intercept these when Console passes them through
    this.cachedLicenseData = {}
    this.cachedSettings = {}

    this.memoryManager = new SharedMemoryManager(config.relay)

    const zmqConfigured = (config.sockets.commandRemoteAddr &&
                           config.sockets.dataRemoteAddr &&
                           config.sockets.bulkRemoteAddr)
    const wsConfigured = !!config.sockets.wsRemoteUrl
    if (!(zmqConfigured || wsConfigured) ||
         (zmqConfigured && wsConfigured)) {
      throw new Error('Relay requires one of ZMQ or Websocket remotes be configured')
    }

    // Data zmq listener
    if (config.sockets.dataBindAddr) {
      this.dataPull = new ZMQBindSocket({
        bindUrl: config.sockets.dataBindAddr,
        privateKey: config.sockets.privateKey,
        hwm: config.sockets.HWM,
        type: 'pull'
      })

      // Relay all `data` messages, queue if necessary.
      this.dataPull.socket.on('message', message => {
        this.handleData(message)
      })
    }

    // Data zmq forwarder
    if (config.sockets.dataRemoteAddr) {
      this.dataPush = new ZMQRemoteSocket({
        remoteUrl: config.sockets.dataRemoteAddr,
        publicKey: config.sockets.remotePublicKey || config.sockets.publicKey,
        hwm: config.sockets.HWM,
        type: 'push'
      })
    }

    // engine.io listener
    if (this.config.sockets.wsBindAddr) {
      this.wsListener = new websocket.Listener(this)
    }

    // engine.io forwarder
    if (this.config.sockets.wsRemoteUrl) {
      this.wsForwarder = new websocket.Forward(this)
    }

    const dataBufferQueue = new BufferQueue({
      name: 'dataPush',
      config: config.relay,
      sendBuffered: buffers => {
        if (this.wsForwarder) {
          this.wsForwarder.sendData(buffers[0])
        }
        if (this.dataPush) {
          this.dataPush.socket.send(buffers)
        }
        if (this.secondaryZmqRemotes) {
          for (let secondaryRemote of this.secondaryZmqRemotes) {
            secondaryRemote.dataPush.socket.send(buffers)
          }
        }
        if (this.secondaryWsRemotes) {
          for (let secondaryRemote of this.secondaryWsRemotes) {
            secondaryRemote.sendData(buffers[0])
          }
        }
      }
    })
    this.dataBufferQueue = dataBufferQueue
    if (this.dataPush) {
      this.dataPush.socket.on('connect', () => {
        this.dataBufferQueue.resume()
      })
      this.dataPush.socket.on('disconnect', () => {
        this.dataBufferQueue.pause()
      })
    }
    if (this.wsForwarder) {
      this.wsForwarder.on('open', () => {
        this.dataBufferQueue.resume()
      })
      this.wsForwarder.on('close', () => {
        this.dataBufferQueue.pause()
      })
    }

    // n.b. Buffering is intentionally not triggered on secondary remotes

    // Register the Data Queue in the memory manager
    this.memoryManager.registerBufferQueue(dataBufferQueue)

    // Bulk zmq listener
    if (config.sockets.bulkBindAddr) {
      this.bulkPull = new ZMQBindSocket({
        bindUrl: config.sockets.bulkBindAddr,
        privateKey: config.sockets.privateKey,
        hwm: config.sockets.bulkHWM,
        type: 'pull'
      })

      // Relay all `bulk` messages, queue if necessary.
      this.bulkPull.socket.on('message', (message, data) => {
        this.handleBulk(message, data)
      })
    }

    // Bulk zmq forwarder
    if (config.sockets.bulkRemoteAddr) {
      this.bulkPush = new ZMQRemoteSocket({
        remoteUrl: config.sockets.bulkRemoteAddr,
        publicKey: config.sockets.remotePublicKey || config.sockets.publicKey,
        hwm: config.sockets.bulkHWM,
        type: 'push'
      })
    }

    const bulkBufferQueue = new BufferQueue({
      name: 'bulkPush',
      config: config.relay,
      sendBuffered: buffers => {
        if (this.wsForwarder) {
          this.wsForwarder.sendBulk(buffers[0], buffers[1])
        }
        if (this.bulkPush) {
          this.bulkPush.socket.send(buffers)
        }
        if (this.secondaryZmqRemotes) {
          for (let secondaryRemote of this.secondaryZmqRemotes) {
            secondaryRemote.bulkPush.socket.send(buffers)
          }
        }
        if (this.secondaryWsRemotes) {
          for (let secondaryRemote of this.secondaryWsRemotes) {
            secondaryRemote.sendBulk(buffers[0], buffers[1])
          }
        }
      }
    })
    this.bulkBufferQueue = bulkBufferQueue
    if (this.bulkPush) {
      this.bulkPush.socket.on('connect', () => {
        this.bulkBufferQueue.resume()
      })
      this.bulkPush.socket.on('disconnect', () => {
        this.bulkBufferQueue.pause()
      })
    }
    if (this.wsForwarder) {
      this.wsForwarder.on('open', () => {
        this.bulkBufferQueue.resume()
      })
      this.wsForwarder.on('close', () => {
        this.bulkBufferQueue.pause()
      })
    }

    // n.b. Buffering is intentionally not triggered on secondary remotes

    // Register the Bulk Queue in the memory manager
    this.memoryManager.registerBufferQueue(bulkBufferQueue)

    // Command zmq publisher
    if (config.sockets.commandBindAddr) {
      this.commandPub = new ZMQBindSocket({
        bindUrl: config.sockets.commandBindAddr,
        privateKey: config.sockets.privateKey,
        hwm: config.sockets.HWM,
        type: 'xpub'
      })
      this.commandPub.socket.on('message', message => {
        // An XPUB message is a subscription.  The first byte in the buffer will be
        // 1 for subscribe and 0 for unsubscribe.  The remaining bytes are the
        // subscription string.
        var isSub = message[0]
        var subscription = message.slice(1).toString()
        if (isSub && /^configuration/.test(subscription)) {
          // If license is missing, request it from upstream.  When it arrives
          // it will cause us to re-send this packet with the updated license,
          // which should be fine.
          if (!this.cachedLicenseData.hasOwnProperty('token')) {
            this.requestConfig()
          }
          this.sendConfig(subscription)
        }
      })
    }

    // Command zmq subscriber
    if (config.sockets.commandRemoteAddr) {
      this.commandSub = new ZMQRemoteSocket({
        remoteUrl: config.sockets.commandRemoteAddr,
        publicKey: config.sockets.remotePublicKey || config.sockets.publicKey,
        hwm: config.sockets.HWM,
        type: 'sub'
      })

      // Relay all `command` messages, no buffering or HWM
      this.commandSub.socket.on('message', (topic, message) => {
        var proxyCommand = true
        try {
          const parsed = JSON.parse(message)

          // config and license payloads should not be proxied directly because
          // they can pass incorrect network connectivity information to agents
          if (!parsed.command) {
            proxyCommand = false

            if (this.commandSub.configSubscribed) {
              this.commandSub.socket.unsubscribe('configuration')
              this.commandSub.configSubscribed = false
            }

            let rebroadcast = false
            if (parsed.license) {
              // This is a config / license payload, save the license for passthrough
              this.cachedLicenseData = parsed.license
              rebroadcast = true
            }
            if (parsed.settings) {
              // Contains settings that the agents should apply, such as blocked loop threshold
              this.cachedSettings = parsed.settings
              rebroadcast = true
            }

            // Re-send the config / license data to downstream connections so
            // they are updated
            if (rebroadcast) {
              this.sendConfig('broadcast')
            }
          }
        } catch (err) {
          Logger.warn('Relaying an unparsable command message!')
        }

        if (proxyCommand) {
          this.handleCommand(topic, message)
        }
      })
    }

    this.createSecondaryRemotes()
  }

  createSecondaryRemotes () {
    this.destroySecondaryRemotes()

    if (this.config.sockets.secondaryRemotes) {
      this.secondaryZmqRemotes = []
      this.secondaryWsRemotes = []
      for (let secondaryRemote of this.config.sockets.secondaryRemotes) {
        if (secondaryRemote.commandRemoteAddr &&
            secondaryRemote.dataRemoteAddr &&
            secondaryRemote.bulkRemoteAddr) {
          // Create all 3 zmq objects for this endpoint
          const commandSub = new ZMQRemoteSocket({
            remoteUrl: secondaryRemote.commandRemoteAddr,
            publicKey: secondaryRemote.remotePublicKey || this.config.sockets.publicKey,
            hwm: this.config.sockets.HWM,
            type: 'sub'
          })

          // Relay all `command` messages, no buffering or HWM
          commandSub.socket.on('message', (topic, message) => {
            var proxyCommand = true
            try {
              const parsed = JSON.parse(message)

              // Don't proxy config and license data from secondary remotes
              if (!parsed.command) {
                proxyCommand = false
              }
            } catch (err) {
              Logger.warn('Relaying an unparsable command message!')
            }

            if (proxyCommand) {
              this.handleCommand(topic, message)
            }
          })
          const dataPush = new ZMQRemoteSocket({
            remoteUrl: secondaryRemote.dataRemoteAddr,
            publicKey: secondaryRemote.remotePublicKey || this.config.sockets.publicKey,
            hwm: this.config.sockets.HWM,
            type: 'push'
          })

          const bulkPush = new ZMQRemoteSocket({
            remoteUrl: secondaryRemote.bulkRemoteAddr,
            publicKey: secondaryRemote.remotePublicKey || this.config.sockets.publicKey,
            hwm: this.config.sockets.bulkHWM,
            type: 'push'
          })

          this.secondaryZmqRemotes.push({ commandSub, dataPush, bulkPush })
        } else if (secondaryRemote.wsRemoteUrl) {
          const wsForwarder = new websocket.Forward(this, secondaryRemote.wsRemoteUrl)
          this.secondaryWsRemotes.push(wsForwarder)
        } else {
          Logger.warn(`Unrecognized secondary remote config: ${JSON.stringify(secondaryRemote)}`)
        }
      }
    }
  }

  destroySecondaryRemotes () {
    if (this.secondaryWsRemotes) {
      for (let secondaryRemote of this.secondaryWsRemotes) {
        secondaryRemote.close()
      }
    }
    if (this.secondaryZmqRemotes) {
      for (let secondaryRemote of this.secondaryZmqRemotes) {
        secondaryRemote.commandSub.close()
        secondaryRemote.dataPush.close()
        secondaryRemote.bulkPush.close()
      }
    }
    this.secondaryWsRemotes = null
    this.secondaryZmqRemotes = null
  }

  logCommand (topic, message) {
    Logger.debug('Relaying a command message on %s -', topic.toString('utf8'), message.toString('utf8'))
  }

  handleCommand (topic, message) {
    if (Logger.getLogLevel() === 'debug') {
      this.logCommand(topic, message)
    }

    if (this.wsListener) {
      this.wsListener.sendCommand(topic, message)
    }
    if (this.commandPub) {
      this.commandPub.socket.send([topic, message])
    }
  }

  logData (msgBuffer) {
    const message = msgBuffer.toString('utf8')
    let parsed
    try {
      parsed = JSON.parse(message)
    } catch (e) {
      Logger.debug('Received a corrupted data message', message)
      return
    }

    if (parsed.command === 'metrics') {
      return // Skip the flood of metrics data
    }

    const status = this.dataBufferQueue.buffering ? 'Buffering' : 'Relaying'
    Logger.debug(`${status} a data message`, parsed.command)
  }

  handleData (message) {
    if (Logger.getLogLevel() === 'debug') {
      this.logData(message)
    }

    if (!this.dataBufferQueue.buffering) {
      if (this.wsForwarder) {
        this.wsForwarder.sendData(message)
      }
      if (this.dataPush) {
        this.dataPush.socket.send(message)
      }
      if (this.secondaryZmqRemotes) {
        for (let secondaryRemote of this.secondaryZmqRemotes) {
          secondaryRemote.dataPush.socket.send(message)
        }
      }
      if (this.secondaryWsRemotes) {
        for (let secondaryRemote of this.secondaryWsRemotes) {
          secondaryRemote.sendData(message)
        }
      }
      return
    }

    this.dataBufferQueue.queueBuffers([message])
  }

  logBulk (msgBuffer) {
    const message = msgBuffer.toString('utf8')
    let parsed
    try {
      parsed = JSON.parse(message)
    } catch (e) {
      Logger.debug('Received a corrupted bulk message', message)
      return
    }
    const status = this.bulkBufferQueue.buffering ? 'Buffering' : 'Relaying'
    Logger.debug(`${status} a bulk message`, parsed.command)
  }

  handleBulk (message, data) {
    if (Logger.getLogLevel() === 'debug') {
      this.logBulk(message)
    }

    if (!this.bulkBufferQueue.buffering) {
      if (this.wsForwarder) {
        this.wsForwarder.sendBulk(message, data)
      }
      if (this.bulkPush) {
        this.bulkPush.socket.send([message, data])
      }
      if (this.secondaryZmqRemotes) {
        for (let secondaryRemote of this.secondaryZmqRemotes) {
          secondaryRemote.bulkPush.socket.send([message, data])
        }
      }
      if (this.secondaryWsRemotes) {
        for (let secondaryRemote of this.secondaryWsRemotes) {
          secondaryRemote.sendBulk(message, data)
        }
      }
      return
    }

    this.bulkBufferQueue.queueBuffers([message, data])
  }

  sendConfig (topic) {
    var payload = {
      config: this.context.config,
      license: this.cachedLicenseData,
      settings: this.cachedSettings
    }
    this.commandPub.socket.send([topic, JSON.stringify(payload)])
  }

  requestConfig () {
    this.commandSub.socket.subscribe('configuration')
    this.commandSub.configSubscribed = true
  }

  reloadConfig (config) {
    Logger.info('Reloading socket config')
    this.config = config

    // Only reload upstream sockets
    if (this.dataPush) {
      this.dataPush.setOptions(config.sockets.dataRemoteAddr, config.sockets.publicKey)
    }
    if (this.bulkPush) {
      this.bulkPush.setOptions(config.sockets.bulkRemoteAddr, config.sockets.publicKey)
    }
    if (this.commandSub) {
      this.commandSub.setOptions(config.sockets.commandRemoteAddr, config.sockets.publicKey)
    }
    if (this.wsForwarder) {
      this.wsForwarder.connect()
    }

    this.createSecondaryRemotes()

    Logger.info('Reloading BufferQueue config')
    this.dataBufferQueue.setOptions(config.relay)
    this.bulkBufferQueue.setOptions(config.relay)
    this.memoryManager.setOptions(config.relay)
  }

  shutdown () {
    Logger.info('Shutting down ZeroMQ')
    this.destroySecondaryRemotes()
    if (this.dataPull) this.dataPull.close()
    if (this.dataPush) this.dataPush.close()
    if (this.bulkPull) this.bulkPull.close()
    if (this.bulkPush) this.bulkPush.close()
    if (this.commandSub) this.commandSub.close()
    if (this.commandPub) this.commandPub.close()
    if (this.wsForwarder) this.wsForwarder.close()
    if (this.wsListener) this.wsListener.close()

    Logger.info('Closing BufferQueues')
    this.bulkBufferQueue.close()
    this.dataBufferQueue.close()
  }
}

module.exports = ZeroMQRelay
