'use strict'

const zmqzap = require('zmq-zap')
const zap = new zmqzap.ZAP()
const util = require('../util')
const ZMQBindSocket = require('./socket')
const lodash = require('lodash')

const Logger = util.getLogger(__dirname, __filename)

zap.use(new zmqzap.CurveMechanism(function (data, callback) {
  // Authorize all ZeroMQ socket connections.
  callback(null, true)
}))

class ZeroMQServer {
  constructor (params) {
    this.config = params.config
    this.context = params.context
    this.bindCb = util.onlyCallOnce(params.bindCb)
    this.preBindCb = params.preBindCb
    this._prevAgentSettings = {}

    var boundCount = 0
    var socksBound = err => {
      if (!this.bindCb) return
      if (err) this.bindCb(err)
      if (++boundCount === 3) this.bindCb()
    }

    this.data = new ZMQBindSocket({
      name: 'data',
      bindUrl: this.config.sockets.dataBindAddr,
      privateKey: this.config.sockets.privateKey,
      hwm: this.config.sockets.HWM,
      type: 'pull',
      bindCb: socksBound,
      preBindCb: this.preBindCb,
      server: this
    })

    this.bulk = new ZMQBindSocket({
      name: 'bulk',
      bindUrl: this.config.sockets.bulkBindAddr,
      privateKey: this.config.sockets.privateKey,
      hwm: this.config.sockets.bulkHWM,
      type: 'pull',
      bindCb: socksBound,
      preBindCb: this.preBindCb,
      server: this
    })

    this.command = new ZMQBindSocket({
      name: 'command',
      bindUrl: this.config.sockets.commandBindAddr,
      privateKey: this.config.sockets.privateKey,
      hwm: this.config.sockets.HWM,
      type: 'xpub',
      bindCb: socksBound,
      preBindCb: this.preBindCb,
      server: this
    })

    if (this.context.eventManager) {
      this.context.eventManager.on('notifications-settings-changed', () => {
        this.settingsChanged()
      })
    } else {
      Logger.warn('Event manager not available.')
    }
  }

  send (message) {
    if (message.id) {
      this.sendToAgent(message)
    } else {
      this.broadcast(message)
    }
  }

  sendToAgent (message) {
    if (!this.command.socket) {
      Logger.info('Message could not be sent after socket closed', message)
    }

    Logger.debug('sending agent command message', message)
    var id = message.id
    message = JSON.stringify(message)
    this.command.socket.send([`id:${id}`, message])
  }

  broadcast (message) {
    if (!this.command.socket) {
      Logger.info('Message could not be sent after socket closed', message)
    }

    Logger.debug('sending broadcast command message', message)
    message = JSON.stringify(message)
    this.command.socket.send(['broadcast', message])
  }

  settingsChanged () {
    this.context.settingsManager.getSettings('notifications', (err, notifs) => {
      if (err) {
        // We still need to send an empty object, so don't return
        Logger.warn('Settings error during agent notification.', err)
      }

      const threshold = lodash.get(notifs, 'nsolid-process-blocked.threshold')
      if (threshold) {
        if (!this._prevAgentSettings.blockedLoopThreshold || this._prevAgentSettings.blockedLoopThreshold !== threshold) {
          // When the threshold value changes in settings,
          // we need to notifiy all the agents.
          this.sendConfig('broadcast')
        } // else, do nothing, the threshold value is set but hasn't changed
      } else {
        // No threshold value is set in the settings
        if (this._prevAgentSettings.blockedLoopThreshold) {
          // We used to have one, so we need to notify the agents
          // that the setting has been deleted.
          this.sendConfig('broadcast')
        } // else no threshold value has been set or sent to agents at all
      }
    })
  }

  sendConfig (topic) {
    if (!this.command.socket) return

    const provider = this.context.provider || {}
    const license = provider.license || {}
    const token = license.token

    // Send our blocked loop threshold value from settings or undefined
    // if it's not set in the settings.
    const sendPayload = (err, settings) => {
      let agentSettings = {}
      if (!err && settings && settings['nsolid-process-blocked']) {
        // Make sure a non-zero number is passed here,
        // otherwise drop it out of the json
        agentSettings.blockedLoopThreshold = Number(settings['nsolid-process-blocked'].threshold) || undefined
      }

      var payload = {
        config: this.context.config,
        license: { token },
        settings: agentSettings
      }

      this.command.socket.send([topic, JSON.stringify(payload)])

      // Track what we've send to _all_ agents so if it changes we can rebroadcast
      if (topic === 'broadcast') {
        this._prevAgentSettings = agentSettings
      }
    }

    if (this.context.settingsManager) {
      this.context.settingsManager.getSettings('notifications', sendPayload)
    } else {
      // If no settings manager is available, we will still send
      // an empty JSON object.
      sendPayload()
    }
  }

  shutdown () {
    Logger.info('Shutting down ZeroMQ')

    closeCatching('data', this.data)
    closeCatching('bulk', this.bulk)
    closeCatching('command', this.command)
  }
}

function closeCatching (name, socket) {
  try {
    socket.close()
  } catch (err) {
    Logger.warn(err, `error closing zmq socket ${name}`)
  }
}

module.exports = ZeroMQServer
