'use strict'

const util = require('../util')
const zmq = util.requireZmq()

const nsolidConsole = require('../nsolid-console')

const Logger = util.getLogger(__dirname, __filename)

class ZMQBindSocket {
  constructor (params) {
    this.bindUrl = params.bindUrl
    this.name = params.name
    this.type = params.type
    this.bindCb = params.bindCb
    this.preBindCb = params.preBindCb
    this.server = params.server

    if (!/:\/\//.test(this.bindUrl)) {
      this.bindUrl = 'tcp://' + this.bindUrl
    }

    if (/^(sub|push)$/.test(this.type)) {
      Logger.error(`Use ZMQRemoteSocket for ${this.type} sockets (${params.bindUrl || params.remoteUrl})`)
      nsolidConsole.shutdown(1)
    }

    this.socket = zmq.socket(params.type)

    this.socket.monitor(500, 0)
    this.socket.setsockopt(zmq.ZMQ_SNDHWM, params.hwm)
    this.socket.setsockopt(zmq.ZMQ_IPV6, 1)

    // Make every subscription fire a message even if it's redundant.  Without
    // this, agents trying to get configuration at the same time will fail.
    if (this.type === 'xpub') {
      this.socket.setsockopt(zmq.ZMQ_XPUB_VERBOSE, 1)
    }

    this.socket.curve_server = 1
    this.socket.curve_secretkey = params.privateKey

    this.socket.on('bind', () => {
      Logger.debug(`${this.type} socket bound to ${this.bindUrl}`)
      if (this.bindCb) {
        this.bindCb()
      }
    })
    this.socket.on('bind_error', (event, endpoint, ex) => this.exitOnError(ex || event))

    if (this.preBindCb) {
      this.preBindCb(this.name, this.socket, this.server)
    }
    this.socket.bind(this.bindUrl)

    this.socket.on('accept', () => Logger.debug(`${this.type} socket accepted a remote connection to ${this.bindUrl}`))
  }

  exitOnError (err) {
    if (!(err instanceof Error)) err = new Error(`zmq bind error: ${err}`)
    Logger.error(err, `Couldn't bind ${this.type} channel socket to ${this.bindUrl}`)
    if (this.bindCb) {
      this.bindCb(err)
    } else {
      nsolidConsole.shutdown(1)
    }
  }

  close () {
    // Already closed
    if (!this.socket) return

    this.socket.setsockopt(zmq.ZMQ_LINGER, 0)
    // If we bound to a wildcard, ZMQ doesn't actually allow you to unbind from
    // that same string, so you must ask it for what it believes the bind
    // address is.  Who even knows why.
    var unbindUrl = this.socket.getsockopt(zmq.ZMQ_LAST_ENDPOINT)
    this.socket.unbind(unbindUrl, err => {
      if (err) {
        Logger.error(`Error while unbinding ZMQ socket: ${err.stack}`)
      }
      this.socket.removeAllListeners()
      this.socket.unmonitor()
      this.socket.close()
      this.socket = null
    })
  }
}

module.exports = ZMQBindSocket
