'use strict'

exports.handleRequest = handleRequest

// exposed for unit testing
exports.getVulnReport = getVulnReport
exports.getUniqueVulns = getUniqueVulns

const util = require('../../util')
const Logger = util.getLogger(__dirname, __filename)

// Handle the HTTP request
function handleRequest (req, res) {
  const context = req.nsConsole.context

  Logger.debug('handling request')

  // get the agents
  const agents = context.agentManager.list()

  // get unique vulns across all agents as map of vulnId -> [ agents ]
  const vulnsMap = getUniqueVulns(agents)

  // build the final report
  const result = []
  for (let vulnId of vulnsMap.keys()) {
    const agents = vulnsMap.get(vulnId)
    result.push(getVulnReport(context, vulnId, agents))
  }

  res.status(200)
  res.send({
    time: new Date().toISOString(),
    vulnerabilities: result
  })
}

function getVulnReport (context, vulnId, agents) {
  const processes = []

  for (let agent of agents) {
    if (agent.vulns == null) continue

    let depsChains = []
    if (agent.packagesHash != null) {
      const packageInfo = context.packageManager.getPackageInfo(agent.packagesHash)
      depsChains = packageInfo.depsChainsMap.get(vulnId) || []
    }

    const topLevelVulns = depsChains.length

    // push the process info to the vuln record
    processes.push({
      id: agent.info.id,
      app: agent.info.app,
      tags: agent.info.tags,
      hostname: agent.info.hostname,
      topLevel: topLevelVulns,
      depChains: depsChains
    })
  }

  const vuln = context.vulnManager.getVulnById(vulnId)
  return {
    report: vuln,
    processes: processes
  }
}

// return a map of vulnId -> [ agents ]
function getUniqueVulns (agents) {
  const vulns = new Map()

  for (let agent of agents) {
    if (agent.vulns == null) continue

    for (let vuln of agent.vulns) {
      const vulnId = vuln.id
      let agentArray = vulns.get(vulnId)
      if (agentArray == null) {
        agentArray = []
        vulns.set(vulnId, agentArray)
      }

      agentArray.push(agent)
    }
  }

  return vulns
}
